from __future__ import annotations


__all__ = [
    'NamedCallableProxy',
]


# -- IMPORTS --

# -- Standard libraries --
from typing import Any, Callable

# -- 3rd party libraries --

# -- Internal libraries --


class NamedCallableProxy:
    """Class wrapper to have named callable proxies, which can also work as :py:class:`enum.Enum` values.

    Adapted from Stack Overflow solution by Ceppo93:

        https://stackoverflow.com/a/40486992
    """
    __slots__ = ('_callable', '_name')

    _callable: Callable
    _name: str

    def __new__(cls, callable_: Callable, /, *, name: str = None) -> NamedCallableProxy:
        """Constructor

        Parameters
        ----------
        callable_ : `callable`
            The callable to name and proxy.

        name : `str`, default=None
            The user-defined name of the callable to use in :py:func:`~continuedfractions.utils.__repr__`.
            If :py:data:`None` the Python-defined default will be used.

        Returns
        -------
        callable
            A named callable proxy.

        Examples
        --------
        >>> square = NamedCallableProxy(lambda x: x ** 2, name="square: x |--> x^2")
        >>> square
        NamedCallableProxy("square: x |--> x^2")
        >>> list(map(square, [1, 2, 3, 4, 5]))
        [1, 4, 9, 16, 25]
        """
        self = super().__new__(cls)
        self._callable = callable_
        self._name = name

        return self

    @property
    def name(self) -> str:
        return self._name

    def __repr__(self) -> str:
        if self._name:
            return f'{self.__class__.__name__}("{self._name}")'

        return str(self._callable)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._callable(*args, **kwargs)


if __name__ == "__main__":      # pragma: no cover
    # Doctest the module from the project root using
    #
    #     PYTHONPATH="src" python3 -m doctest -v src/continuedfractions/utils.py
    #
    import doctest
    doctest.testmod()



from __future__ import annotations


__all__ = [
    'rationals',
    'coprime_pairs',
    'FareyFraction',
    'farey_sequence',
    'KSRMTree',
]


# -- IMPORTS --

# -- Standard libraries --
import math
import typing

from itertools import chain

# -- 3rd party libraries --

# -- Internal libraries --
from continuedfractions.utils import NamedCallableProxy
from continuedfractions.continuedfraction import ContinuedFraction


def rationals(
    enumeration: typing.Literal["cantor diagonal", "cantor diagonal transposed", "rectilinear", "rectilinear transposed"],
    positive_only: bool = True
) -> typing.Generator[ContinuedFraction, None, None]:
    """:py:class:`typing.Generator` : An enumerator of rational numbers :math:`\\mathbb{Q}` using an enumeration on Cantor's 2D representation of the rationals.

    By default it generates sequences of all positive rational numbers as
    :py:class:`~continuedfractions.continuedfractions.ContinuedFraction`
    objects with the user-specified enumeration on Cantor's 2D
    representation of the rationals.

    To include negative rationals and :math:`0` set ``positive_only`` to
    ``False``, as described in the documentation and the docstring examples
    below.

    The enumeration type should be given by the ``enumeration`` argument, which
    should be one of the following:

    * ``"cantor diagonal"``: for the standard way of enumerating rational
    numbers using Cantor diagonalisation (with ``positive_only=False``):
    
    .. math::

        \\frac{1}{1}, \\frac{2}{1}, \\frac{1}{2}, \\frac{1}{3}, \\frac{3}{1}, \\frac{4}{1}, \\frac{3}{2}, \\frac{2}{3}, \\frac{1}{4}, \\ldots

    * ``"cantor diagonal transposed"``: similar to the standard Cantor
    diagonalisation (with ``positive_only=False``), except the enumeration
    proceeds:
    ::

        right 1 step ->
        diagonal down 1 step ->
        down 1 step ->
        diagonal up 2 steps ->
        ...
    
    and produces the sequence:

    .. math::

        \\frac{1}{1}, \\frac{1}{2}, \\frac{2}{1}, \\frac{3}{1}, \\frac{1}{3}, \\frac{1}{4}, \\frac{2}{3}, \\frac{3}{2}, \\frac{4}{1}, \\ldots

    * ``"rectilinear"``: an enumeration which proceeds in a reverse L-shaped
    way (with ``positive_only=False``):
    ::

        down 1 step  ->
        right 1 step -> up 1 step   -> right 1 step ->
        down 2 steps -> left 2 steps -> down 1 step ->
        right 3 steps -> up 3 steps -> right 1 step ->
        down 4 steps  -> left 4 steps -> down 1 step ->
        ...

    and produces the sequence: 
    
    .. math::

       \\frac{1}{1}, \\frac{2}{1}, \\frac{1}{2}, \\frac{1}{3}, \\frac{2}{3}, \\frac{3}{2}, \\frac{3}{1}, \\frac{4}{1}, \\ldots

    * ``"rectilinear transposed"``: an enumeration similar to ``"rectilinear"``
    (with ``positive_only=False``) except the enumeration proceeds:
    ::

        right 1 step  ->
        down 1 step -> left 1 step   -> down 1 step ->
        right 2 steps -> up 2 steps -> right 1 step ->
        down 3 steps -> left 3 steps -> down 1 step ->
        right 4 steps  -> up 4 steps -> right 1 step ->
        ...
    
    and produces the sequence:
    
    .. math::

       \\frac{1}{1}, \\frac{1}{2}, \\frac{2}{1}, \\frac{3}{1}, \\frac{3}{2}, \\frac{2}{3}, \\frac{1}{3}, \\frac{1}{4}, \\ldots

    Parameters
    ----------
    enumeration : str
        A string literal which describes how the rationals should be
        enumerated:

        * ``"cantor diagonal"``
        * ``"cantor diagonal transposed"``
        * ``"rectilinear"``
        * ``"rectilinear transposed"``

    positive_only : bool, default=True
        Whether to generate only the positive rationals, which is true by
        default.

    Yields
    ------
    ContinuedFraction
        The rational numbers are generated as
        :py:class:`~continuedfractions.continuedfraction.ContinuedFraction`
        objects.

    Raises
    ------
    ValueError
        If ``enumeration`` is not an unrecognised or unsupported enumeration.

    Examples
    --------
    >>> rats = rationals(enumeration="cantor diagonal")
    >>> first_ten = [next(rats) for _ in range(10)]
    >>> first_ten
    [ContinuedFraction(1, 1), ContinuedFraction(2, 1), ContinuedFraction(1, 2), ContinuedFraction(1, 3), ContinuedFraction(3, 1), ContinuedFraction(4, 1), ContinuedFraction(3, 2), ContinuedFraction(2, 3), ContinuedFraction(1, 4), ContinuedFraction(1, 5)]
    >>> rats = rationals(enumeration="cantor diagonal", positive_only=False)
    >>> first_ten_plus_zero = [next(rats) for _ in range(11)]
    >>> first_ten_plus_zero
    [ContinuedFraction(0, 1), ContinuedFraction(1, 1), ContinuedFraction(-1, 1), ContinuedFraction(2, 1), ContinuedFraction(-2, 1), ContinuedFraction(1, 2), ContinuedFraction(-1, 2), ContinuedFraction(1, 3), ContinuedFraction(-1, 3), ContinuedFraction(3, 1), ContinuedFraction(-3, 1)]
    >>> rats = rationals(enumeration="cantor diagonal transposed")
    >>> first_ten = [next(rats) for _ in range(10)]
    >>> first_ten
    [ContinuedFraction(1, 1), ContinuedFraction(1, 2), ContinuedFraction(2, 1), ContinuedFraction(3, 1), ContinuedFraction(1, 3), ContinuedFraction(1, 4), ContinuedFraction(2, 3), ContinuedFraction(3, 2), ContinuedFraction(4, 1), ContinuedFraction(5, 1)]
    >>> rats = rationals(enumeration="cantor diagonal transposed", positive_only=False)
    >>> first_ten_plus_zero = [next(rats) for _ in range(11)]
    >>> first_ten_plus_zero
    [ContinuedFraction(0, 1), ContinuedFraction(1, 1), ContinuedFraction(-1, 1), ContinuedFraction(1, 2), ContinuedFraction(-1, 2), ContinuedFraction(2, 1), ContinuedFraction(-2, 1), ContinuedFraction(3, 1), ContinuedFraction(-3, 1), ContinuedFraction(1, 3), ContinuedFraction(-1, 3)]
    >>> rats = rationals(enumeration="rectilinear")
    >>> first_ten = [next(rats) for _ in range(10)]
    >>> first_ten
    [ContinuedFraction(1, 1), ContinuedFraction(2, 1), ContinuedFraction(1, 2), ContinuedFraction(1, 3), ContinuedFraction(2, 3), ContinuedFraction(3, 2), ContinuedFraction(3, 1), ContinuedFraction(4, 1), ContinuedFraction(4, 3), ContinuedFraction(3, 4)]
    >>> rats = rationals(enumeration="rectilinear", positive_only=False)
    >>> first_ten_plus_zero = [next(rats) for _ in range(11)]
    >>> first_ten_plus_zero
    [ContinuedFraction(0, 1), ContinuedFraction(1, 1), ContinuedFraction(-1, 1), ContinuedFraction(2, 1), ContinuedFraction(-2, 1), ContinuedFraction(1, 2), ContinuedFraction(-1, 2), ContinuedFraction(1, 3), ContinuedFraction(-1, 3), ContinuedFraction(2, 3), ContinuedFraction(-2, 3)]
    >>> rats = rationals(enumeration="rectilinear transposed", positive_only=False)
    >>> first_ten_plus_zero = [next(rats) for _ in range(11)]
    >>> first_ten_plus_zero
    [ContinuedFraction(0, 1), ContinuedFraction(1, 1), ContinuedFraction(-1, 1), ContinuedFraction(1, 2), ContinuedFraction(-1, 2), ContinuedFraction(2, 1), ContinuedFraction(-2, 1), ContinuedFraction(3, 1), ContinuedFraction(-3, 1), ContinuedFraction(3, 2), ContinuedFraction(-3, 2)]
    >>> rats = rationals(enumeration="some unknown enumeration")
    >>> next(rats)
    Traceback (most recent call last):
    ...
    ValueError: The value of `enumeration` should be one of the following: "cantor diagonal", "cantor diagonal transposed", "rectilinear" or "rectilinear transposed".
    """
    # Validation of the ``enumeration`` argument.
    enumeration_ = enumeration.strip().lower()
    if enumeration_ not in [
        "cantor diagonal", "cantor diagonal transposed", "rectilinear", "rectilinear transposed"
    ]:
        raise ValueError(
            'The value of `enumeration` should be one of the following: '
            '"cantor diagonal", "cantor diagonal transposed", "rectilinear" '
            'or "rectilinear transposed".'
        )

    if enumeration_ == "cantor diagonal" and positive_only:
        # The diagonal index, starting at ``n = 1``: the first diagonal `D1`
        # which is just ``1/1``.
        n = 1
        yield ContinuedFraction(1, 1)

        while True:
            # Increment the diagonal index, and generate the new diagonal:
            # the diagonal ``D_n`` is given by:
            #
            #     n/1, (n - 1)/2, (n - 2)/3, ... , 1/n
            #
            # if ``n`` is even, while if ``n`` is odd it is given by:
            #
            #    1/n, 2/(n - 1), 3/(n - 2), ... , n/1
            #
            n += 1
            us = range(n, 0, -1) if n % 2 == 0 else range(1, n + 1)
            vs = range(1, n + 1) if n % 2 == 0 else range(n, 0, -1)
            yield from (ContinuedFraction(u, v) for u, v in zip(us, vs) if math.gcd(u, v) == 1)
    elif enumeration_ == "cantor diagonal" and not positive_only:
        # When ``positive_only`` is ``False`` we start by generating the
        # ``ContinuedFraction`` object for ``0``.
        yield ContinuedFraction(0, 1)
        n = 0

        while True:
            # Same process as for ``"cantor diagonal"``, except that for each
            # coprime pair ``(u, v)`` on the ``n``-th diagonal we generate the
            # negative continued fraction ``-u/v`` immediately after the
            # positive continued fraction ``u/v``.
            n += 1
            us = range(n, 0, -1) if n % 2 == 0 else range(1, n + 1)
            vs = range(1, n + 1) if n % 2 == 0 else range(n, 0, -1)
            pos_rationals = (ContinuedFraction(u, v) for u, v in zip(us, vs) if math.gcd(u, v) == 1)
            yield from chain.from_iterable(((r, -r) for r in pos_rationals))
    elif enumeration_ == "cantor diagonal transposed" and positive_only:
        # The diagonal index, starting at ``n = 1``: the first diagonal `D1`
        # which is just ``1/1``.
        n = 1
        yield ContinuedFraction(1, 1)

        while True:
            # Increment the diagonal index, and generate the new diagonal:
            # the diagonal ``D_n`` is given by:
            #     1/n, 2/(n - 1), 3/(n - 2), ... , n/1  
            # 
            # if ``n`` is even, while if ``n`` is odd it is given by:
            #
            #     n/1, (n - 1)/2, (n - 2)/3, ... , 1/n
            #
            n += 1
            us = range(1, n + 1) if n % 2 == 0 else range(n, 0, -1)
            vs = range(n, 0, -1) if n % 2 == 0 else range(1, n + 1)
            yield from (ContinuedFraction(u, v) for u, v in zip(us, vs) if math.gcd(u, v) == 1)
    elif enumeration_ == "cantor diagonal transposed" and not positive_only:
        # When ``positive_only`` is ``False`` we start by generating the
        # ``ContinuedFraction`` object for ``0``.
        yield ContinuedFraction(0, 1)
        n = 0

        while True:
            # Same process as for ``"cantor diagonal transposed"``, except that
            # for each coprime pair ``(u, v)`` on the ``n``-th diagonal we
            # generate the negative continued fraction ``-u/v`` immediately
            # after the positive continued fraction ``u/v``.
            n += 1
            us = range(1, n + 1) if n % 2 == 0 else range(n, 0, -1)
            vs = range(n, 0, -1) if n % 2 == 0 else range(1, n + 1)
            pos_rationals = (ContinuedFraction(u, v) for u, v in zip(us, vs) if math.gcd(u, v) == 1)
            yield from chain.from_iterable(((r, -r) for r in pos_rationals))
    elif enumeration_ == "rectilinear" and positive_only:
        # The initial reverse-L index, starting at ``n = 1``: this will
        # generate ``1/1``.
        n = 1
        yield ContinuedFraction(1, 1)

        while True:
            # Increment the reverse-L index, and enumerate the rationals (as
            # described in the documentation) from the  reverse Ls ``⅃_n`` first
            # by first chaining the subsequences ``⅃_{n,1}`` and ``⅃_{n,2}``,
            # both of length ``2n - 1``, to produce the chained subsequence:
            #
            #     n/1, n/2, n/3, ... , n/n, (n - 1)/n, (n - 2)/n, (n - 3)/n, ... , 1/n
            # 
            # if ``n`` is even, or, if ``n`` is odd, as:
            #
            #     1/n, 2/n, 3/n, ... , n/n, n/(n - 1), n/(n - 2), n/(n - 3), ... , n/1
            #
            # and then filtering out composite fractions ``u/v`` in the chained
            # subsequence, i.e. applying ``math.gcd(u, v) == 1``. 
            #
            n += 1
            us = (
                chain((n for _ in range(n)), range(n - 1, 0, -1))
                if n % 2 == 0 else 
                chain(range(1, n + 1), (n for _ in range(n - 1)))
            )
            vs = (
                chain(range(1, n + 1), (n for _ in range(n - 1)))
                if n % 2 == 0 else
                chain((n for _ in range(n)), range(n - 1, 0, -1))
            )
            yield from (ContinuedFraction(u, v) for u, v in zip(us, vs) if math.gcd(u, v) == 1)
    elif enumeration_ == "rectilinear" and not positive_only:
        # When ``positive_only`` is ``False`` we start by generating the
        # ``ContinuedFraction`` object for ``0``.
        yield ContinuedFraction(0, 1)
        n = 0

        while True:
            # Same process as for ``"rectilinear"``, except that for each
            # coprime pair ``(u, v)`` on the ``n``-th reverse L ``⅃_n`` we
            # generate the negative continued fraction ``-u/v`` immediately
            # after the positive continued fraction ``u/v``.
            n += 1
            us = (
                chain((n for _ in range(n)), range(n - 1, 0, -1))
                if n % 2 == 0 else 
                chain(range(1, n + 1), (n for _ in range(n - 1)))
            )
            vs = (
                chain(range(1, n + 1), (n for _ in range(n - 1)))
                if n % 2 == 0 else
                chain((n for _ in range(n)), range(n - 1, 0, -1))
            )
            pos_rationals = (ContinuedFraction(u, v) for u, v in zip(us, vs) if math.gcd(u, v) == 1)
            yield from chain.from_iterable(((r, -r) for r in pos_rationals))
    elif enumeration_ == "rectilinear transposed" and positive_only:
        # The initial reverse-L index, starting at ``n = 1``: this will
        # generate ``1/1``.
        n = 1
        yield ContinuedFraction(1, 1)

        while True:
            # Increment the reverse-L index, and enumerate the rationals (as
            # described in the documentation) from the  reverse Ls ``⅃_n`` first
            # by first chaining the subsequences ``⅃_{n,1}`` and ``⅃_{n,2}``,
            # both of length ``2n - 1``, to produce the chained subsequence:
            #
            #     1/n, 2/n, 3/n, ... , n/n, n/(n - 1), n/(n - 2), n/(n - 3), ... , n/1
            # 
            # if ``n`` is even, or, if ``n`` is odd, as:
            #
            #     n/1, n/2, n/3, ... , n/n, (n - 1)/n, (n - 2)/n, (n - 3)/n, ... , 1/n
            #
            # and then filtering out composite fractions ``u/v`` in the chained
            # subsequence, i.e. applying ``math.gcd(u, v) == 1``.
            n += 1
            us = (
                chain(range(1, n), (n for _ in range(n)))
                if n % 2 == 0 else
                chain((n for _ in range(n)), range(n - 1, 0, -1))
            )
            vs = (
                chain((n for _ in range(n - 1)), range(n, 0, -1))
                if n % 2 == 0 else
                chain(range(1, n + 1), (n for _ in range(n - 1)))
            )
            yield from (ContinuedFraction(u, v) for u, v in zip(us, vs) if math.gcd(u, v) == 1)
    elif enumeration_ == "rectilinear transposed" and not positive_only:
        # When ``positive_only`` is ``False`` we start by generating the
        # ``ContinuedFraction`` object for ``0``.
        yield ContinuedFraction(0, 1)
        n = 0

        while True:
            # Same process as for ``"rectilinear transposed"``, except that
            # for each coprime pair ``(u, v)`` on the ``n``-th reverse L
            # ``⅃_n`` we generate the negative continued fraction ``-u/v``
            # immediately after the positive continued fraction ``u/v``. 
            n += 1
            us = (
                chain(range(1, n), (n for _ in range(n)))
                if n % 2 == 0 else
                chain((n for _ in range(n)), range(n - 1, 0, -1))
            )
            vs = (
                chain((n for _ in range(n - 1)), range(n, 0, -1))
                if n % 2 == 0 else
                chain(range(1, n + 1), (n for _ in range(n - 1)))
            )
            pos_rationals = (ContinuedFraction(u, v) for u, v in zip(us, vs) if math.gcd(u, v) == 1)
            yield from chain.from_iterable(((r, -r) for r in pos_rationals))

    return  # pragma: no cover


def _coprime_integers(n: int, /) -> typing.Generator[int, None, None]:
    """Generates a sequence of (positive) integers :math:`1 \\leq m < n` coprime to a given positive integer :math:`n`.

    The tuple is sorted in descending order of magnitude.

    Parameters
    ----------
    n : int
        The positive integer for which (positive) coprime integers
        :math:`m < n` are sought.

    Raises
    ------
    ValueError
        If :math:`n` is either not a positive integer, or :math:`n > 1` such
        that either the ``start`` value is **not** in the range
        :math:`1..n - 1` or the ``stop`` value is **not** in the range
        :math:`\\text{start} + 1..n`.

    Yields
    ------
    int
        A sequence of (positive) integers :math:`1 \\leq m < n` coprime to a
        given positive integer :math:`n`.

    Examples
    --------
    Examples using the default ``start`` and ``stop`` values:

    >>> tuple(_coprime_integers(1))
    (1,)
    >>> tuple(_coprime_integers(5))
    (4, 3, 2, 1)
    >>> tuple(_coprime_integers(10))
    (9, 7, 3, 1)

    """
    if not isinstance(n, int) or n < 1:
        raise ValueError(
            "`n` must be a positive integer; if `n` > 1 then the "
            "`start` value must be a positive integer in the range 1..n - 1; "
            "and if given the `stop` value must be a positive integer in the "
            "range `start` + 1..n"
        )

    yield from (m for m in range(n, 0, -1) if math.gcd(m, n) == 1)


KSRMTreeNode: typing.TypeAlias = tuple[int, int]        # Custom type for nodes of the KSRM coprime pairs tree
KSRMTreeBranch: typing.TypeAlias = NamedCallableProxy   # Custom type for generating branches of the KSRM coprime pairs tree


class KSRMTree:
    """A class implementation of the Kanga-Saunders-Randall-Mitchell (KSRM) ternary trees for representing and generating pairs of all (positive) coprime integers.

    The term "KSRM trees" is the author's, and refers to the trees presented in the following papers:

    * Kanga, A. R. (1990). The Family Tree of Pythagorean Triplets. The Mathematical Gazette, 26(15), 15-17.
    * Mitchell, D. W. (2001). An Alternative Characterisation of All Primitive Pythagorean Triples. The Mathematical Gazette, 85(503), 273-275. https://doi.org/10.2307/3622017
    * Saunders, R., & Randall, T. (1994). The family tree of the Pythagorean triplets revisited. The Mathematical Gazette, 78(482), 190-193. https://doi.org/10.2307/3618576

    .. note::

       The class is named ``KSRMTree`` purely for convenience, but it is
       actually a representation of two (ternary) subtrees.

    .. note::

       The author could not access the Kanga paper online, but the core result
       is described clearly in the papers of Saunders and Randall, and of
       Mitchell.

    """

    # The two slots for the private attributes for the key tree features - the
    # two roots ``(2, 1), (3, 1)``, and the three branch generating functions.
    __slots__ = ('_roots', '_branches')

    _roots: tuple[KSRMTreeNode, KSRMTreeNode]

    _branches: tuple[KSRMTreeBranch]    # noqa: F821

    def __new__(cls) -> KSRMTree:
        """Class constructor.

        Creates and initialises a new instance with the minimum set of required
        attributes - for the roots and branch generating functions, which we
        call branches.
        """
        self = super().__new__(cls)
        self._roots = ((2, 1), (3, 1))
        self._branches = (
            NamedCallableProxy(lambda x, y: (2 * x - y, x), name="KSRM tree branch #1: (x, y) |--> (2x - y, x)"),
            NamedCallableProxy(lambda x, y: (2 * x + y, x), name="KSRM tree branch #2: (x, y) |--> (2x + y, x)"),
            NamedCallableProxy(lambda x, y: (x + 2 * y, y), name="KSRM tree branch #3: (x, y) |--> (x + 2y, y)")
        )

        return self

    @property
    def roots(self) -> typing.Literal[((2, 1), (3, 1))]:
        """:py:class:`tuple`: The tuple of roots of the KSRM trees, which are :math:`(2, 1)` and :math:`(3, 1)`.

        For more details see the following papers:

        * Kanga, A. R. (1990). The Family Tree of Pythagorean Triplets. The Mathematical Gazette, 26(15), 15-17.
        * Mitchell, D. W. (2001). An Alternative Characterisation of All Primitive Pythagorean Triples. The Mathematical Gazette, 85(503), 273-275. https://doi.org/10.2307/3622017
        * Saunders, R., & Randall, T. (1994). The family tree of the Pythagorean triplets revisited. The Mathematical Gazette, 78(482), 190-193. https://doi.org/10.2307/3618576

        Examples
        --------
        >>> KSRMTree().roots
        ((2, 1), (3, 1))
        """
        return self._roots

    @property
    def branches(self) -> tuple[KSRMTreeBranch]:    # noqa: F821
        """:py:class:`tuple`: The tuple of three branch generating functions of the KSRM trees.

        There are three branch generating functions, given by the mappings:

        .. math::

           \\begin{align}
           (a, b) &\\longmapsto (2a - b, a) \\\\
           (a, b) &\\longmapsto (2a + b, a) \\\\
           (a, b) &\\longmapsto (a + 2b, b)
           \\end{align}

        For more details see the following papers:

        * Kanga, A. R. (1990). The Family Tree of Pythagorean Triplets. The Mathematical Gazette, 26(15), 15-17.
        * Mitchell, D. W. (2001). An Alternative Characterisation of All Primitive Pythagorean Triples. The Mathematical Gazette, 85(503), 273-275. https://doi.org/10.2307/3622017
        * Saunders, R., & Randall, T. (1994). The family tree of the Pythagorean triplets revisited. The Mathematical Gazette, 78(482), 190-193. https://doi.org/10.2307/3618576

        Examples
        --------
        Generating the first two generations of children for the parent
        :math:`(2, 1)`.

        >>> tree = KSRMTree()
        >>> tree.branches
        (NamedCallableProxy("KSRM tree branch #1: (x, y) |--> (2x - y, x)"), NamedCallableProxy("KSRM tree branch #2: (x, y) |--> (2x + y, x)"), NamedCallableProxy("KSRM tree branch #3: (x, y) |--> (x + 2y, y)"))
        >>> tree.branches[0](2, 1)
        (3, 2)
        >>> tree.branches[1](2, 1)
        (5, 2)
        >>> tree.branches[2](2, 1)
        (4, 1)
        >>> tree.branches[0](*tree.branches[0](2, 1))
        (4, 3)
        >>> tree.branches[1](*tree.branches[0](2, 1))
        (8, 3)
        >>> tree.branches[2](*tree.branches[0](2, 1))
        (7, 2)
        >>> tree.branches[0](*tree.branches[1](2, 1))
        (8, 5)
        >>> tree.branches[1](*tree.branches[1](2, 1))
        (12, 5)
        >>> tree.branches[2](*tree.branches[1](2, 1))
        (9, 2)
        >>> tree.branches[0](*tree.branches[2](2, 1))
        (7, 4)
        >>> tree.branches[1](*tree.branches[2](2, 1))
        (9, 4)
        >>> tree.branches[2](*tree.branches[2](2, 1))
        (6, 1)
        """
        return self._branches

    def _backtrack(
        self,
        n: int,
        visited: list[tuple[KSRMTreeNode, KSRMTreeBranch]],     # noqa: F821
        /,
        *,
        node_bound: int = None
    ) -> tuple[KSRMTreeNode, KSRMTreeBranch, int, KSRMTreeBranch]:  # noqa: F821
        """Backtracks on the KSRM coprime pairs trees from a failed node to the nearest previously visited node that satisfies the node bound.

        A private function that backtracks on the KSRM coprime pairs trees: the
        procedure is that, given a (positive) integer :math:`n > 2`, for which
        coprime pairs are being sought, and a sequence (list) of pairs of
        visited nodes and their associated generating branches in the KSRM
        tree, and assuming that the last element of the visited sequence
        contains the node that "failed", the function identifies the nearest
        previously visited node whose first component satisifes the test
        :math:`< n` **and** and whose associated generating branch is not equal
        to the third branch given by :math:`(x, y) \\longmapsto (x + 2y, y)`.

        .. note::

           The function assumes that the last node in the incoming sequence
           of visited nodes and generating branch pairs represents a "failed"
           node, i.e. whose first component failed the test :math:`\\leq n`
           during the search. No attempt is made to validate or verify the
           failed node, and the only purpose of the function is to backtrack
           to the nearest previously visited node which meets the requirements
           listed above.

        .. note::

           There is no input validation as this is a private function which
           will be called from
           :py:meth:`~continuedfractions.sequences.KSRMTree.search_root`. So
           results for invalid arguments will most likely be incorrect or
           unexpected.

        Parameters
        ----------
        n : int
            The (positive) integer :math:`> 2` which is passed by the root
            search method or the general tree search method.

        visited : list
            A sequence of visited nodes and associated generating branches in
            the KSRM coprime pairs tree.

        node_bound : int, default=None
            A bound to check that :math:`a < n` for a node :math:`(a, b)`. The
            actual default value is the incoming :math:`n`, and this is set
            internally.

        Returns
        -------
        tuple
            A tuple consisting of the following values in order: (1) the
            target node in the visited sequence to backtrack to, (2) the
            associated generating branch function (the lambda for branch #1,
            or branch #2, or branch #3), (3) the index of the target node
            and branch pair in the visited sequence, (4) the generating
            branch of the successor node of the target node returned as (1).

        Examples
        --------
        An example where :math:`n = 5` and the failed node is :math:`(6, 1)`,
        which was the successor node to :math:`(4, 1)` from the third branch.

        >>> tree = KSRMTree()
        >>> tree.branches
        (NamedCallableProxy("KSRM tree branch #1: (x, y) |--> (2x - y, x)"), NamedCallableProxy("KSRM tree branch #2: (x, y) |--> (2x + y, x)"), NamedCallableProxy("KSRM tree branch #3: (x, y) |--> (x + 2y, y)"))
        >>> visited = [((2, 1), None), ((4, 1), tree.branches[-1]), ((6, 1), tree.branches[-1])]
        >>> tree._backtrack(5, visited)
        ((2, 1), None, 0, NamedCallableProxy("KSRM tree branch #3: (x, y) |--> (x + 2y, y)"))

        An example where :math:`n = 8` and the failed node is :math:`(19, 8)`,
        which was the successor node to :math:`(8, 3)` from the first branch.

        >>> tree = KSRMTree()
        >>> tree.branches
        (NamedCallableProxy("KSRM tree branch #1: (x, y) |--> (2x - y, x)"), NamedCallableProxy("KSRM tree branch #2: (x, y) |--> (2x + y, x)"), NamedCallableProxy("KSRM tree branch #3: (x, y) |--> (x + 2y, y)"))
        >>> visited = [((2, 1), None), ((3, 2),tree.branches[0]), ((8, 3),tree.branches[1]), ((19, 8), tree.branches[0])]
        >>> tree._backtrack(8, visited)
        ((3, 2), NamedCallableProxy("KSRM tree branch #1: (x, y) |--> (2x - y, x)"), 1, NamedCallableProxy("KSRM tree branch #2: (x, y) |--> (2x + y, x)"))
        """
        # Set the node bound for ``r``: so we require ``a < n`` for the
        # backtracked target node.
        node_bound = node_bound or n

        # Set the current node and branch as the last pair in the visited
        # sequence.
        cur_node, cur_branch = visited[-1]

        # If we've only visited one node it must be the root, and there is
        # no further backtracking possible, so just return appropriately.
        if len(visited) == 1:
            return cur_node, cur_branch, 0, None

        # Otherwise do some initialisation for variables tracking the
        # generating branch for the last visited node, and the current
        # node index.
        last_branch = None
        cur_index = len(visited) - 1

        # The main backtracking loop - while there are more nodes to backtrack
        # to, go back one node, decrement the current node index and set the
        # current node and generating branch, and also set the generating
        # branch of the last visited node before the current node.
        #
        # If the current node passes the test ``a < n` and we are not on the
        # last branch, return the current node, generating branch, index
        # and the generating branch of the last visited node before the current
        # node.
        while cur_index > 0 and (cur_node[0] >= node_bound or last_branch == self.branches[-1]):
            cur_index -= 1
            cur_node, cur_branch = visited[cur_index]
            last_branch = visited[cur_index + 1][1]

        # Return the current node, generating branch, index and the generating
        # branch of the last visited node before the current node.
        return cur_node, cur_branch, cur_index, last_branch

    def search_root(self, n: int, root: KSRMTreeNode, /) -> typing.Generator[KSRMTreeNode, None, None]:
        """Depth-first branch-and-bound generative search function (in pre-order, NLMR), with backtracking and pruning, on the KSRM coprime pairs trees, starting from the given root node.

        The given root node need not be the canonical roots, :math:`(2, 1)`,
        :math:`(3, 1)`, but can be any of their successor nodes.

        It is required that :math:`n \\geq 2`, otherwise a
        :py:class:`ValueError` is raised.

        The search implementation is an iterative version of a depth-first
        branch-and-bound search (DFS) procedure, with backtracking and pruning,
        in which nodes are traversed in NLMR pre-order (root -> left -> mid ->
        right) and bounds and checks are applied to the nodes, including
        pruning failed or unnecessary nodes, before further traversal or
        backtracking:
        
        #. Visit the current node :math:`(a, b)` and check :math:`a \\leq n`.
        #. If the check is successful iteratively traverse the current node's
           first child and its children, then the second child and its 
           children, and then the third child and its children, applying the
           check :math:`a \\leq n` to each visited node.
        #. If a check fails for any node backtrack to the nearest previously
           visited node which meets a stricter check :math:`a < n` and which
           has unvisited child nodes, while pruning all visited intermediate
           nodes after the backtracked target node and leading up to the failed
           node, including the failed node. By design the backtracked target
           node will have untraversed children on at least one branch, and the
           traversal can begin again, as described above.

        Parameters
        ----------
        n : int
            The positive integer for which coprime pairs :math:`(a, b)`, with
            :math:`1 \\leq b < a \\leq n`, are sought.

        root : tuple
            The "root" node from which to search - this can be either of the
            canonical roots, :math:`(2, 1)`, :math:`(3, 1)`, but also any of
            their successor nodes.

        Raises
        ------
        ValueError
            If ``n`` is not an integer or is :math:`< 2`.

        Yields
        ------
        tuple
            Pairs of coprime integers :math:`(a, b)`, with
            :math:`1 \\leq b < a \\leq n`.

        Examples
        --------
        Searching from the root :math:`(2, 1)` for coprime pairs for
        :math:`n = 5`:

        >>> tree = KSRMTree()
        >>> list(tree.search_root(5, (2, 1)))
        [(2, 1), (3, 2), (4, 3), (5, 4), (5, 2), (4, 1)]
        >>> assert tree.roots[0] == (2, 1)
        >>> list(tree.search_root(5, tree.roots[0]))
        [(2, 1), (3, 2), (4, 3), (5, 4), (5, 2), (4, 1)]

        The same type of search from the root :math:`(3, 1)`:

        >>> list(tree.search_root(5, (3, 1)))
        [(3, 1), (5, 3), (5, 1)]
        >>> assert tree.roots[1] == (3, 1)
        >>> list(tree.search_root(5, tree.roots[1]))
        [(3, 1), (5, 3), (5, 1)]
        """
        # Input validation.
        if not isinstance(n, int) or n <= 1 or not math.gcd(*root) == 1:
            raise ValueError(
                "`n` must be a positive integer >= 2, and `root` must be a "
                "coprime pair (r, s) for integers r > s >= 1"
            )

        if n < root[0]:
            return

        # A stack to store visited nodes and their generating branches.
        visited = []

        # A counter to store the number of nodes searched (or visited) -
        # useful for debugging and also for optimising the search
        # implementation.
        num_nodes_searched = 0

        # Start at the root, initialising variables for the current node
        # and generating branch to that of the root node, and also 
        # initialising a variable to store the generating branch of the
        # successor node of the current node.
        cur_node = root
        cur_branch = last_branch = None
        visited.append((cur_node, cur_branch))
        num_nodes_searched += 1

        # Generate the root
        yield cur_node
                    
        # The iterative backtracking depth-first branch-and-bound search
        # (pre-order, LNMR), with pruning of intermediate and failed nodes.
        while True:
            # If starting from either root, which do not have generating
            # branches, set the current branch to branch #1.
            if cur_branch is None:
                cur_branch = self.branches[0]

            # Generate and visit the next node ``(a, b)``, where ``1 <= b < a``
            # and ``gcd(a, b) = 1`` is guaranteed by the nature of the
            # generating branches.
            cur_node = cur_branch(*cur_node)
            visited.append((cur_node, cur_branch))
            num_nodes_searched += 1

            # If the node satisfies ``a <= n`` and generate it, then update the
            # variable storing the generating branch of the successor node of
            # the current node, and set the current (next generating) branch to
            # branch #1, and continue the DFS.
            #
            # If the node does not satisfy ``a <= n`` backtrack to the nearest
            # satisfying non-root node, prune any unnecessary nodes as needed,
            # and continue the DFS on unexplored branches. If there are no
            # nearest satisfying non-root nodes and all remaining branches have
            # been explored the DFS has ended, and so exit.
            if cur_node[0] <= n:
                yield cur_node
                last_branch = cur_branch
                cur_branch = self.branches[0]
                continue
            else:
                # Backtrack to the nearest satisfying target node, which will
                # become the current node; the current branch and current
                # node index are also updated, as is the variable storing the
                # generating branch of the successor node of the target/current
                # node.
                cur_node, cur_branch, cur_index, last_branch = self._backtrack(n, visited, node_bound=n)

                # Prune all visited intermediate nodes after the backtracked
                # target node leading up to the failed node, including the
                # failed node.
                visited[cur_index + 1:] = []

                # If we've reached the root node, and it has no untraversed
                # children, then we've finished our DFS, so return.
                if cur_node == root and last_branch == self.branches[-1]:
                    return

                # Otherwise, switch to the generating branch of the "next"
                # child node - branch #2 if the current branch is branch #1, or
                # branch #3 (the last branch) if the current branch is #2 - and
                # continue the DFS.
                cur_branch = self.branches[1] if last_branch.name == self.branches[0].name else self.branches[-1]
                continue

        # Not strictly required, but this has been inserted to make
        # debugging easier.
        return

    def search(self, n: int, /) -> typing.Generator[KSRMTreeNode, None, None]:
        """Depth-first branch-and-bound generative search function (in pre-order, NLMR) on the KSRM coprime pairs trees to find all pairs of coprime (positive) integers not exceeding the given integer :math:`n \\geq 1`.
    
        See the :py:meth:`~continuedfractions.sequences.KSRMTree.search_root`
        method for details of the implementation for the root-based search.

        This method mainly calls the root-based search method
        :py:meth:`~continuedfractions.sequences.KSRMTree.search_root` for the
        two canonical roots :math:`(2, 1)` and :math:`(3, 1)`.

        The two KSRM trees are actually only traversed if :math:`n > 3`.

        Parameters
        ----------
        n : int
            The positive integer for which coprime pairs :math:`(a, b)`, with
            :math:`1 \\leq b < a \\leq n`, are sought.

        Raises
        ------
        ValueError
            If ``n`` is not an integer or is :math:`< 1`.

        Yields
        ------
        tuple
            Pairs of coprime integers :math:`(a, b)`, with
            :math:`1 \\leq b < a \\leq n`.

        Examples
        --------
        A few examples of invalid and valid searches:

        >>> tree = KSRMTree()
        >>> list(tree.search("not an integer"))
        Traceback (most recent call last):
        ...
        ValueError: `n` must be a positive integer.
        >>> list(tree.search(1))
        [(1, 1)]
        >>> list(tree.search(2))
        [(1, 1), (2, 1)]
        >>> list(tree.search(3))
        [(1, 1), (2, 1), (3, 2), (3, 1)]
        >>> list(tree.search(5))
        [(1, 1), (2, 1), (3, 2), (4, 3), (4, 1), (3, 1), (5, 4), (5, 3), (5, 2), (5, 1)]
        >>> list(tree.search(10))
        [(1, 1), (2, 1), (3, 2), (4, 3), (5, 4), (6, 5), (7, 6), (8, 7), (9, 8), (8, 3), (7, 2), (5, 2), (8, 5), (9, 2), (4, 1), (7, 4), (9, 4), (6, 1), (8, 1), (3, 1), (5, 3), (7, 5), (9, 7), (7, 3), (5, 1), (9, 5), (7, 1), (9, 1), (10, 9), (10, 7), (10, 3), (10, 1)]
        """
        if not isinstance(n, int) or n < 1:
            raise ValueError("`n` must be a positive integer.")

        if n == 1:
            yield (1, 1)
        if n == 2:
            yield from ((1, 1), (2, 1))
        elif n == 3:
            yield from ((1, 1), (2, 1), (3, 2), (3, 1))
        elif n > 3:
            yield from chain(
                ((1, 1),),
                self.search_root(n - 1, self.roots[0]),
                self.search_root(n - 1, self.roots[1]),
                ((n, i) for i in _coprime_integers(n))
            )


def coprime_pairs(n: int, /) -> typing.Generator[tuple[int, int], None, None]:
    """Generates a sequence (tuple) of all pairs :math:`(a, b)` of (positive) coprime integers :math:`a, b` such that :math:`a, b <= n`.

    It does **not** call on the
    :py:meth:`~continuedfractions.sequences.KSRMTree.search` method, which is
    not as fast as this one.

    A :py:class:`ValueError` is raised if ``n`` is not an :py:class:`int`
    or is an :py:class:`int` less than :math:`1`.

    Parameters
    ----------
    n : int
        The positive integer for which coprime pairs :math:`(a, b)`, with
        :math:`1 \\leq b < a \\leq n`, are sought.

    Raises
    ------
    ValueError
        If ``n`` is not an :py:class:`int` or is an :py:class:`int` less than
        :math:`1`.

    Yields
    ------
    tuple
        A :py:class:`tuple` of pairs of coprime integers :math:`(a, b)`, with
        :math:`1 \\leq b < a \\leq n`.

    Examples
    --------
    A few examples of coprime pairs generation:

    >>> tuple(coprime_pairs(1))
    ((1, 1),)
    >>> tuple(coprime_pairs(2))
    ((1, 1), (2, 1))
    >>> tuple(coprime_pairs(3))
    ((1, 1), (2, 1), (3, 1), (3, 2))
    >>> tuple(coprime_pairs(5))
    ((1, 1), (2, 1), (3, 1), (3, 2), (4, 1), (4, 3), (5, 1), (5, 2), (5, 3), (5, 4))
    >>> tuple(coprime_pairs(10))
    ((1, 1), (2, 1), (3, 1), (3, 2), (4, 1), (4, 3), (5, 1), (5, 2), (5, 3), (5, 4), (6, 1), (6, 5), (7, 1), (7, 2), (7, 3), (7, 4), (7, 5), (7, 6), (8, 1), (8, 3), (8, 5), (8, 7), (9, 1), (9, 2), (9, 4), (9, 5), (9, 7), (9, 8), (10, 1), (10, 3), (10, 7), (10, 9))
    >>> tuple(coprime_pairs("invalid input"))
    Traceback (most recent call last):
    ...
    ValueError: `n` must be a positive integer
    """
    if not isinstance(n, int) or n < 1:
        raise ValueError("`n` must be a positive integer")

    yield (1, 1)

    for i in range(1, n + 1):
        for j in (_j for _j in range(1, i) if _j < i and math.gcd(i, _j) == 1):
            yield (i, j)


class FareyFraction(ContinuedFraction):
    """A simple wrapper class for Farey fractions, subclassing :py:class:`~continuedfractions.continuedfraction.ContinuedFraction`.
    """


def farey_sequence(n: int, /) -> typing.Generator[FareyFraction, None, None]:
    """Generates an (ordered) sequence (tuple) of rational numbers forming the Farey sequence of order :math:`n`.

    The elements of the sequence are yielded as
    :py:class:`~continuedfractions.sequences.FareyFraction`
    instances, in ascending order of magnitude.

    Parameters
    ----------
    n : int:
        The order of the Farey sequence.

    Raises
    ------
    ValueError
        If :math:`n` is not a positive integer.

    Yields
    ------
    FareyFraction
        A sequence of
        :py:class:`~continuedfractions.sequences.FareyFraction`
        instances representing the elements of the Farey sequence of order
        :math:`n`, in ascending order of magnitude.

    Examples
    --------
    >>> tuple(farey_sequence(1))
    (FareyFraction(0, 1), FareyFraction(1, 1))
    >>> tuple(farey_sequence(2))
    (FareyFraction(0, 1), FareyFraction(1, 2), FareyFraction(1, 1))
    >>> tuple(farey_sequence(3))
    (FareyFraction(0, 1), FareyFraction(1, 3), FareyFraction(1, 2), FareyFraction(2, 3), FareyFraction(1, 1))
    >>> tuple(farey_sequence(4))
    (FareyFraction(0, 1), FareyFraction(1, 4), FareyFraction(1, 3), FareyFraction(1, 2), FareyFraction(2, 3), FareyFraction(3, 4), FareyFraction(1, 1))
    >>> tuple(farey_sequence(5))
    (FareyFraction(0, 1), FareyFraction(1, 5), FareyFraction(1, 4), FareyFraction(1, 3), FareyFraction(2, 5), FareyFraction(1, 2), FareyFraction(3, 5), FareyFraction(2, 3), FareyFraction(3, 4), FareyFraction(4, 5), FareyFraction(1, 1))
    """
    if not isinstance(n, int) or n < 1:
        raise ValueError("`n` must be a positive integer")

    if n == 1:
        yield from (FareyFraction(0, 1), FareyFraction(1, 1))
    else:
        yield FareyFraction(0, 1)
        yield from sorted(FareyFraction(q, p) for (p, q) in coprime_pairs(n))


if __name__ == "__main__":      # pragma: no cover
    # Doctest the module from the project root using
    #
    #     PYTHONPATH="src" python3 -m doctest -v src/continuedfractions/sequences.py
    #
    import doctest
    doctest.testmod()



__all__ = [
    'continued_fraction_real',
    'continued_fraction_rational',
    'convergent',
    'convergents',
    'fraction_from_coefficients',
    'fraction_from_elements',
    'left_mediant',
    'mediant',
    'remainder',
    'remainders',
    'right_mediant',
]


# -- IMPORTS --

# -- Standard libraries --
import decimal
import functools
import typing
import warnings

from decimal import Decimal
from fractions import Fraction

# -- 3rd party libraries --

# -- Internal libraries --


def continued_fraction_rational(frac: Fraction, /) -> typing.Generator[int, None, None]:     # pragma: no cover
    """Implementation of the core continued fraction algorithm which generates the (ordered) sequence of coefficients of the (finite) simple continued fraction of the given rational number.

    The resulting sequence of coefficients :math:`a_0,a_1,\\ldots a_n`, defines
    the simple continued fraction:

    .. math::

       a_0 + \\cfrac{1}{a_1 + \\cfrac{1}{a_2 + \\ddots\\cfrac{1}{a_{n - 1} + \\cfrac{1}{a_n}}}}

    which is also written more compactly as:

    .. math::

       [a_0; a_1, a_2\\ldots, a_n]

    The order of the continued fraction is said to be :math:`n`. If the last
    coefficient :math:`a_n = 1` the sequence can be rewritten as
    :math:`[a_0; a_1, a_2\\ldots, a_{n - 1} + 1]`, which is then unique for
    the given rational.

    Negative rational numbers can also be represented in this way, provided we
    use the `Euclidean division lemma <https://en.wikipedia.org/wiki/Euclid%27s_lemma>`_.

    Parameters
    ----------
    frac : `fractions.Fraction`
        The rational number to represented as a simple continued fraction.

    Yields
    ------
    int
        Coefficients of the unique simple continued fraction of the given 
        rational number.

    Examples
    --------
    A few examples are given below of how this function can be used.

    >>> for e in continued_fraction_rational(Fraction(649, 200)):
    ...     print(e)
    ... 
    3
    4
    12
    4
    >>> tuple(continued_fraction_rational(Fraction(415, 93)))
    (4, 2, 6, 7)
    >>> tuple(continued_fraction_rational(Fraction(-649, 200)))
    (-4, 1, 3, 12, 4)
    >>> tuple(continued_fraction_rational(Fraction(123235, 334505)))
    (0, 2, 1, 2, 1, 1, 250, 1, 13)

    Notes
    -----
    Every rational number has a unique simple continued fraction
    :math:`[a_0;a_1,a_2,\\ldots, a_n]`, where :math:`a_n > 1`, and this
    function will always produce this unique represention for any given
    rational.
    """
    x, y = frac.as_integer_ratio()

    while y:
        quo, rem = divmod(x, y)
        yield quo
        x, y = y, rem


def continued_fraction_real(x: int | Fraction | float | str | Decimal, /) -> typing.Generator[int, None, None]:
    """Generates a (finite) sequence of elements of a (simple) continued fraction of the given real number.

    The result may not always be exact, but an approximation depending on
    whether the  given number :math:`x` has an exact binary floating point
    representation.

    The simple continued fraction representation of :math:`x` is a number of
    the form

    .. math::

       a_0 + \\cfrac{1}{a_1 + \\cfrac{1}{a_2 + \\ddots}}

    where :math:`a_0 = [x]` is the integer part of :math:`x`, and the
    :math:`a_1,a_2\\ldots` are the (non-negative) quotients obtained by a
    repeated application of `Euclidean division <https://en.wikipedia.org/wiki/Euclidean_division>`_
    to the fractional part :math:`x - [x]`, which is called the remainder.

    If the last coefficient :math:`a_n = 1` the sequence can be rewritten as
    :math:`[a_0; a_1, a_2\\ldots, a_{n - 1} + 1]`.

    As Python :py:class:`float` values, like all floating point
    implementations, are `finite precision representations <https://docs.python.org/3/tutorial/floatingpoint.html>`_
    of real numbers, the resulting simple continued fraction  of :math:`x`
    generated by this function may be approximate, not exact, and also not
    necessarily unique.

    For non-rational real numbers it is best to pass :py:class:`decimal.Decimal`
    values, with the `context precision <https://docs.python.org/3.12/library/decimal.html#context-objects>`_
    set to the highest level possible.

    The results for rational numbers are guaranteed to be exact however large
    the number, subject to memory and hardware limitations of the running
    environment.

    Invalid values will generate an error in either the
    :py:class:`fractions.Fraction` or :py:class:`decimal.Decimal` classes -
    no errors are raised directly in the function itself.

    Parameters
    ----------
    x : int, fractions.Fraction, float, str, decimal.Decimal
        The real number to represent as a simple continued fraction.

    Yields
    ------
    int
        Elements of a simple continued fraction of the given real number.

    Examples
    --------
    A few examples are given below of how this function can be used.

    >>> list(continued_fraction_real(5000))
    [5000]
    >>> list(continued_fraction_real(Fraction(7, 4)))
    [1, 1, 3]
    >>> list(continued_fraction_real(-5000.0))
    [-5000]
    >>> list(continued_fraction_real(2/5))
    [0, 2, 2, 1801439850948198]
    >>> list(continued_fraction_real('2/5'))
    [0, 2, 2]
    >>> list(continued_fraction_real('-1/3'))
    [-1, 1, 2]
    >>> list(continued_fraction_real(1/1j))
    Traceback (most recent call last):
    ...
    TypeError: conversion from complex to Decimal is not supported
    >>> list(continued_fraction_real("not a numeric string"))
    Traceback (most recent call last):
    ...
    decimal.InvalidOperation: [<class 'decimal.ConversionSyntax'>]
    >>> list(continued_fraction_real(-649/200))
    [-4, 1, 3, 12, 3, 1, 234562480591, 2, 5, 2]
    >>> list(continued_fraction_real('-649/200'))
    [-4, 1, 3, 12, 4]
    >>> list(continued_fraction_real('-649/-200'))
    Traceback (most recent call last):
    ...
    ValueError: Invalid literal for Fraction: '-649/-200'
    >>> list(continued_fraction_real(Decimal('0.3333')))
    [0, 3, 3333]
    """
    if isinstance(x, int):
        yield x
    elif isinstance(x, Fraction):
        yield from continued_fraction_rational(x)
    elif isinstance(x, float):
        yield from continued_fraction_rational(Fraction(*Decimal.from_float(x).as_integer_ratio()))
    elif isinstance(x, str) and '/' in x:
        yield from continued_fraction_rational(Fraction(x))
    else:
        yield from continued_fraction_rational(Fraction(*(Decimal(x).as_integer_ratio())))


def convergent(k: int, *coeffs: int) -> Fraction:
    """Returns the :math:`k`-th convergent of a (simple) continued fraction from a sequence of its coefficients.

    Given a (simple) continued fraction  :math:`[a_0;a_1,a_2,\\ldots]` the
    :math:`k`-th convergent is defined as:

    .. math::

       C_k = a_0 + \\cfrac{1}{a_1 + \\cfrac{1}{a_2 \\ddots \\cfrac{1}{a_{k-1} + \\cfrac{1}{a_k}}}}

    The result is a :py:class:`fractions.Fraction` instance.
    
    The integer :math:`k` is called the order of the convergent, and if 
    :math:`[a_0;a_1,a_2,\\ldots]` is finite of order :math:`n` then it has
    exactly :math:`n + 1` convergents :math:`C_0,C_1,C_2,\\ldots,C_n` where
    the :math:`k`-th convergent :math:`C_k = \\frac{p_k}{q_k}` is given by
    the recurrence relation:

    .. math::
       
       \\begin{align}
       p_k &= a_kp_{k - 1} + p_{k - 2} \\\\
       q_k &= a_kq_{k - 1} + q_{k - 2},        \\hskip{3em}    k \\geq 3
       \\end{align}

    where :math:`p_0 = a_0`, :math:`q_0 = 1`, :math:`p_1 = p_1p_0 + 1`,
    and :math:`q_1 = p_1`.

    This function is a faithful implementation of this algorithm.

    A :py:class:`ValueError` is raised if :math:`k` is not an integer or is an
    integer greater than the number of coefficients, or if any of the
    coefficient are not integers.

    Parameters
    ----------
    k : `int`
        The order of the convergent. Must be a non-negative integer less than
        the number of coefficients.

    *coeffs : `int`
        A variable-length sequence of integer coefficients of a continued
        fraction.

    Returns
    -------
    fractions.Fraction
        A rational fraction constructed from the given sequence of coefficients
        of
        a continued fraction, representing the :math:`k`-order convergent of a
        (finite) simple continued fraction as given by a sequence of coefficients.

    Raises
    ------
    ValueError
        If :math:`k` is not a non-negative integer less than the number of
        coefficients, or if any of the elements are not integers.

    Examples
    --------
    >>> convergent(0, 3, 4, 12, 4)
    Fraction(3, 1)
    >>> convergent(1, 3, 4, 12, 4)
    Fraction(13, 4)
    >>> convergent(2, 3, 4, 12, 4)
    Fraction(159, 49)
    >>> convergent(3, 3, 4, 12, 4)
    Fraction(649, 200)
    >>> convergent(3)
    Traceback (most recent call last):
    ...
    ValueError: `k` must be a non-negative integer not exceeding the order of the continued fraction (number of tail coefficients), and the tail coefficients must all be positive integers.
    >>> convergent(-1, 3, 4, 12, 4)
    Traceback (most recent call last):
    ...
    ValueError: `k` must be a non-negative integer not exceeding the order of the continued fraction (number of tail coefficients), and the tail coefficients must all be positive integers.
    >>> convergent(4, 3, 4, 12, 4)
    Traceback (most recent call last):
    ...
    ValueError: `k` must be a non-negative integer not exceeding the order of the continued fraction (number of tail coefficients), and the tail coefficients must all be positive integers.
    """
    # Define the order of the continued fraction - may be ``-1`` if no elements
    # are given.
    n = len(coeffs) - 1

    if n == -1 or not isinstance(k, int) or k < 0 or k > n or any(not isinstance(coeffs[i], int) or coeffs[i] < 1 for i in range(1, n + 1)):
        raise ValueError(
            "`k` must be a non-negative integer not exceeding the order of "
            "the continued fraction (number of tail coefficients), and the tail "
            "coefficients must all be positive integers."
        )

    a, b = coeffs[0], 1
    
    if k == 0:
        return Fraction(a, b)

    c, d = (coeffs[1] * a) + b, coeffs[1]

    if k == 1:
        return Fraction(c, d)

    for e in coeffs[2:k + 1]:
        p, q = (e * c) + a, (e * d) + b
        a, b = c, d
        c, d = p, q

    return Fraction(p, q)


def convergents(*coeffs: int) -> typing.Generator[Fraction, None, None]:
    """Generates an (ordered) sequence of all convergents of a (simple) continued fraction from a sequence of its coefficients.

    If :math:`n` is the order of the continued fraction represented by the
    given sequence of its coefficients then there are :math:`n + 1` convergents
    :math:`C_0, C_1, \\ldots, C_n`, and the function generates these in that
    order.

    Parameters
    ----------
    *elements : `int`
        A variable-length sequence of integer coefficients of a (simple)
        continued fraction.

    Yields
    ------
    fractions.Fraction
        Each convergent that is generated is a :py:class:`fractions.Fraction`
        instance and a :math:`k`-th convergent of the given continued fraction.

    Raises
    ------
    ValueError
        If there are any non-integer coefficients, or the tail coefficients
        are not positive integers.

    Examples
    --------
    >>> tuple(convergents(3))
    (Fraction(3, 1),)
    >>> tuple(convergents(3, 2))
    (Fraction(3, 1), Fraction(7, 2))
    >>> tuple(convergents(3, 4, 12, 4))
    (Fraction(3, 1), Fraction(13, 4), Fraction(159, 49), Fraction(649, 200))
    >>> tuple(convergents(-5, 1, 1, 6, 7))
    (Fraction(-5, 1), Fraction(-4, 1), Fraction(-9, 2), Fraction(-58, 13), Fraction(-415, 93))
    >>> tuple(convergents(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
    (Fraction(1, 1), Fraction(3, 2), Fraction(10, 7), Fraction(43, 30), Fraction(225, 157), Fraction(1393, 972), Fraction(9976, 6961), Fraction(81201, 56660), Fraction(740785, 516901), Fraction(7489051, 5225670))
    """
    # Define the order of the continued fraction - may be ``-1`` if no coefficients
    # are given.
    n = len(coeffs) - 1

    if n == -1 or any(not isinstance(coeffs[i], int) or (coeffs[i] <= 0 and i > 0) for i in range(n + 1)):
        raise ValueError(
            "Continued fraction elements must be coefficients, and all \n"
            "tail coefficients (from the 1st coefficient onwards) must be positive."
        )

    a, b = coeffs[0], 1
    yield Fraction(a, b)

    if n > 0:
        c, d = (coeffs[1] * a) + b, coeffs[1]
        yield Fraction(c, d)

        if n > 1:
            for e in coeffs[2:]:
                p, q = (e * c) + a, (e * d) + b
                yield Fraction(p, q)
                a, b = c, d
                c, d = p, q


def fraction_from_elements(*elements: int) -> Fraction:
    """Returns the rational number represented by a (simple) continued fraction from a sequence of its elements.

    The elements must be given as positional arguments: if your elements
    are contained in an iterable unpack them in the function call using the
    unpacking operator ``*``, as described in the examples below.

    Parameters
    ----------
    *elements : `int`
        A variable-length sequence of integer elements of a simple continued
        fraction.

    Returns
    -------
    fractions.Fraction
        A rational number constructed from a sequence of elements of a simple
        continued fraction which represents the number.

    Raises
    ------
    ValueError
        If any of the elements are not integers.

    Examples
    --------
    >>> fraction_from_elements(3, 4, 12, 4)
    Fraction(649, 200)
    >>> fraction_from_elements(-4, 1, 3, 12, 4)
    Fraction(-649, 200)
    >>> fraction_from_elements(4, 2, 6, 7)
    Fraction(415, 93)
    >>> fraction_from_elements(*[4, 2, 6, 7])
    Fraction(415, 93)
    >>> fraction_from_elements(4.5, 2, 6, 7)
    Traceback (most recent call last):
    ...
    ValueError: Continued fraction elements must be integers
    """
    warnings.warn(
        "The `from_elements` function is deprecated and will be removed "
        "in future releases. Please use the `from_coefficients` instead."
    )
    if any(not isinstance(elem, int) for elem in elements):
        raise ValueError("Continued fraction elements must be integers")

    return convergent(len(elements) - 1, *elements)


def fraction_from_coefficients(*coeffs: int) -> Fraction:
    """Returns the rational number represented by a (simple) continued fraction from a sequence of its coefficients.

    The coefficient must be given as positional arguments: if your coefficients
    are contained in an iterable unpack them in the function call using the
    unpacking operator ``*``, as described in the examples below.

    Parameters
    ----------
    *coeffs : `int`
        A variable-length sequence of integer coefficients of a simple
        continued fraction.

    Returns
    -------
    fractions.Fraction
        A rational number constructed from a sequence of coefficients of a
        simple continued fraction which represents the number.

    Raises
    ------
    ValueError
        If any of the coefficients are not integers, or the tail coefficients
        are not positive.

    Examples
    --------
    >>> fraction_from_coefficients(3, 4, 12, 4)
    Fraction(649, 200)
    >>> fraction_from_coefficients(-4, 1, 3, 12, 4)
    Fraction(-649, 200)
    >>> fraction_from_coefficients(4, 2, 6, 7)
    Fraction(415, 93)
    >>> fraction_from_coefficients(*[4, 2, 6, 7])
    Fraction(415, 93)
    >>> fraction_from_coefficients(4.5, 2, 6, 7)
    Traceback (most recent call last):
    ...
    ValueError: Continued fraction coefficients must be integers, and all coefficients from the 1st onwards must be positive.
    """
    n = len(coeffs)

    if any(not isinstance(coeffs[i], int) or (i > 0 and coeffs[i] < 1) for i in range(n)):
        raise ValueError(
            "Continued fraction coefficients must be integers, and "
            "all coefficients from the 1st onwards must be positive."
        )

    return convergent(len(coeffs) - 1, *coeffs)


def remainder(k: int, *coeffs: int) -> Fraction:
    """Returns the :math:`k`-th remainder of a (simple) continued fraction from a sequence of its coefficients.

    Given a (simple) continued fraction  :math:`[a_0;a_1,a_2,\\ldots]` the
    :math:`k`-th remainder :math:`R_k` is the (simple) continued fraction
    :math:`[a_k; a_{k + 1}, a_{k + 2}, \\ldots]`:

    .. math::

       R_k = a_k + \\cfrac{1}{a_{k + 1} + \\cfrac{1}{a_{k + 2} \\ddots }}

    where :math:`R_0` is just the original continued fraction, i.e.
    :math:`R_0 = [a_0; a_1, a_2, \\ldots]`.

    The remainders have the property that:

    .. math::

       R_{k - 1} = a_{k - 1} + \\frac{1}{R_k}, \\hskip{3em} k \\geq 1

    If the continued fraction :math:`[a_0; a_1, a_2,\\ldots]` is finite of
    order :math:`n` then all :math:`R_k` are rational. If we let
    :math:`R_k = \\frac{s_k}{t_k}` then the property above can be written as:

    .. math::

       R_{k - 1} = \\frac{s_{k - 1}}{t_{k - 1}} = \\frac{a_{k - 1}s_k + t_k}{s_k}, \\hskip{3em} k \\geq 1

    As this library can only represent finite continued fractions, this
    function always produces remainders as instances of
    :py:class:`~fractions.Fraction`.

    The integer :math:`k` must be non-negative and cannot exceed the order
    of the continued fraction, i.e. the number of tail coefficients.

    Parameters
    ----------
    k : `int`
        The index of the remainder. Must be a non-negative integer not
        exceeding the order of the continued fraction.

    *elements : `int`
        A variable-length sequence of integer coefficients of a (simple)
        continued fraction.

    Returns
    -------
    fractions.Fraction
        A rational fraction constructed from the given sequence of
        coefficients of a continued fraction, representing its :math:`k`-th
        remainder, as defined above.

    Raises
    ------
    ValueError
        If :math:`k` is not an integer, or is an integer greater than the
        number of coefficients, or if any of the coefficients are not integers,
        or if the tail coefficients are not positive integers.

    Examples
    --------
    >>> remainder(0, 3, 4, 12, 4)
    Fraction(649, 200)
    >>> remainder(1, 3, 4, 12, 4)
    Fraction(200, 49)
    >>> remainder(2, 3, 4, 12, 4)
    Fraction(49, 4)
    >>> remainder(3, 3, 4, 12, 4)
    Fraction(4, 1)
    >>> remainder(0, -5, 1, 1, 6, 7)
    Fraction(-415, 93)
    >>> remainder(1, -5, 1, 1, 6, 7)
    Fraction(93, 50)
    >>> remainder(2, -5, 1, 1, 6, 7)
    Fraction(50, 43)
    >>> remainder(3, -5, 1, 1, 6, 7)
    Fraction(43, 7)
    >>> remainder(4, -5, 1, 1, 6, 7)
    Fraction(7, 1)
    >>> remainder(1)
    Traceback (most recent call last):
    ...
    ValueError: `k` must be a non-negative integer not exceeding the order of the continued fraction (number of tail coefficients), and the tail coefficients must all be positive integers.
    >>> remainder(-1, 3, 4, 12, 4)
    Traceback (most recent call last):
    ...
    ValueError: `k` must be a non-negative integer not exceeding the order of the continued fraction (number of tail coefficients), and the tail coefficients must all be positive integers.
    >>> remainder(4, 3, 4, 12, 4)
    Traceback (most recent call last):
    ...
    ValueError: `k` must be a non-negative integer not exceeding the order of the continued fraction (number of tail coefficients), and the tail coefficients must all be positive integers.
    >>> remainder(1, 3, 0, 12, 4)
    Traceback (most recent call last):
    ...
    ValueError: `k` must be a non-negative integer not exceeding the order of the continued fraction (number of tail coefficients), and the tail coefficients must all be positive integers.
    >>> remainder(1, 3, -1, 12, 4)
    Traceback (most recent call last):
    ...
    ValueError: `k` must be a non-negative integer not exceeding the order of the continued fraction (number of tail coefficients), and the tail coefficients must all be positive integers.
    """
    # Define the order of the continued fraction - may be ``-1`` if no
    # coefficients are given.
    n = len(coeffs) - 1

    if n == -1 or not isinstance(k, int) or k < 0 or k > n or any(not isinstance(coeffs[i], int) or coeffs[i] < 1 for i in range(1, n + 1)):
        raise ValueError(
            "`k` must be a non-negative integer not exceeding the order of "
            "the continued fraction (number of tail coefficients), and the tail "
            "coefficients must all be positive integers."
        )

    if k == n:
        return Fraction(coeffs[-1], 1)

    return fraction_from_coefficients(*coeffs[k:])


def remainders(*coeffs: int) -> typing.Generator[Fraction, None, None]:
    """Generates an (ordered) sequence of all remainders of a (simple) continued fraction from a sequence of its coefficients in descending order of index.

    If :math:`n` is the order of the continued fraction represented by the
    given sequence of its coefficients then there are :math:`n + 1` remainders
    :math:`R_0, R_1, \\ldots, R_n`, and the function generates these in
    reverse order :math:`R_0, R_1, \\ldots, R_n`.

    Parameters
    ----------
    *elements : `int`
        A variable-length sequence of integer coefficients of a (simple)
        continued fraction.

    Yields
    ------
    fractions.Fraction
        Each remainder generated is a :py:class:`fractions.Fraction` instance and
        a :math:`k`-th remainder of the given continued fraction.

    Raises
    ------
    ValueError
        If no coefficients are given, or there are any non-integer elements, or
        the tail coefficients are not positive integers.

    Examples
    --------
    >>> tuple(remainders(3))
    (Fraction(3, 1),)
    >>> tuple(remainders(3, 2))
    (Fraction(2, 1), Fraction(7, 2))
    >>> tuple(remainders(3, 4, 12, 4))
    (Fraction(4, 1), Fraction(49, 4), Fraction(200, 49), Fraction(649, 200))
    >>> tuple(remainders(-5, 1, 1, 6, 7))
    (Fraction(7, 1), Fraction(43, 7), Fraction(50, 43), Fraction(93, 50), Fraction(-415, 93))
    >>> tuple(remainders(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
    (Fraction(10, 1), Fraction(91, 10), Fraction(738, 91), Fraction(5257, 738), Fraction(32280, 5257), Fraction(166657, 32280), Fraction(698908, 166657), Fraction(2263381, 698908), Fraction(5225670, 2263381), Fraction(7489051, 5225670))
    >>> tuple(remainders())
    Traceback (most recent call last):
    ...
    ValueError: Continued fraction elements must be integers, and all tail elements (from the 1st element onwards) must be positive.
    >>> tuple(remainders(0, 0))
    Traceback (most recent call last):
    ...
    ValueError: Continued fraction elements must be integers, and all tail elements (from the 1st element onwards) must be positive.
    >>> tuple(remainders(1, 0))
    Traceback (most recent call last):
    ...
    ValueError: Continued fraction elements must be integers, and all tail elements (from the 1st element onwards) must be positive.
    >>> tuple(remainders(1, 2, -1))
    Traceback (most recent call last):
    ...
    ValueError: Continued fraction elements must be integers, and all tail elements (from the 1st element onwards) must be positive.
    """
    # Define the order of the continued fraction - may be ``-1`` if no elements
    # are given.
    n = len(coeffs) - 1

    if n == -1 or any(not isinstance(coeffs[i], int) or (coeffs[i] <= 0 and i > 0) for i in range(n + 1)):
        raise ValueError(
            "Continued fraction elements must be integers, and all "
            "tail elements (from the 1st element onwards) must be positive."
        )

    a, b = coeffs[-1], 1
    yield Fraction(a, b)

    if n > 0:
        i = n - 1

        while i >= 0:
            a, b = coeffs[i] * a + b, a
            yield Fraction(a, b)
            i -= 1


def mediant(r: Fraction, s: Fraction, /, *, dir: str = 'right', k: int = 1) -> Fraction:
    """Returns the :math:`k`-th left- or right-mediant of two rational numbers.

    For a positive integer :math:`k`, the :math:`k`-th left-mediant of two
    rational numbers :math:`r = \\frac{a}{b}` and :math:`s = \\frac{c}{d}`,
    where :math:`b, d, b + d \\neq 0`, is defined as:
    
    .. math::

       \\frac{ka + c}{kb + d}, \\hskip{3em}    k \\geq 1

    while the :math:`k`-th right mediant is defined as:
    
    .. math::

       \\frac{a + kc}{b + kd}, \\hskip{3em}    k \\geq 1

    If we assume that :math:`r < s` and :math:`bd > 0` then these mediants
    have the property that:
   
    .. math::

       \\frac{a}{b} < \\frac{ka + c}{kb + d} \\leq \\frac{a + kc}{b + kd} < \\frac{c}{d},   \\hskip{3em} k \\geq 1

    where equality holds for :math:`k = 1`. If we let :math:`k \\to \\infty`
    then the mediants converge to opposite limits:

    .. math::

      \\begin{align}
      \\lim_{k \\to \\infty} \\frac{ka + c}{kb + d} &= \\frac{a}{b} \\\\
      \\lim_{k \\to \\infty} \\frac{a + kc}{b + kd} &= \\frac{c}{d}
      \\end{align}

    For the left mediant use ``dir="left"``, while for the right use
    ``dir="right"``. The default is ``dir="right"``. For ``k = 1`` the left and
    right mediants are identical to the simple mediant :math:`\\frac{a + c}{b + d}`.

    Parameters
    ----------
    r : `fractions.Fraction`
        The first rational number.

    s : `fractions.Fraction`
        The second rational number.

    dir : `str`, default='right'
        The "direction" of the mediant - `'left'` or `'right'`, as defined
        above.

    k : `int`, default=1
        The order of the mediant, as defined above.

    Returns
    -------
    fractions.Fraction
        The `k`-th left- or right-mediant of the two given rational numbers.

    Examples
    --------
    >>> mediant(Fraction(1, 2), Fraction(3, 5))
    Fraction(4, 7)
    >>> mediant(Fraction(1, 2), Fraction(3, 5), dir='left')
    Fraction(4, 7)
    >>> mediant(Fraction(1, 2), Fraction(3, 5), k=2)
    Fraction(7, 12)
    >>> mediant(Fraction(1, 2), Fraction(3, 5), dir='left', k=2)
    Fraction(5, 9)
    >>> mediant(Fraction(1, 2), Fraction(3, 5), k=3, dir='right')
    Fraction(10, 17)
    >>> mediant(Fraction(1, 2), Fraction(3, 5), k=3, dir='left')
    Fraction(6, 11)
    """
    if not (dir in ['left', 'right'] and isinstance(k, int) and k > 0):
        raise ValueError(
            "The mediant direction must be 'left' or 'right' and the order "
            "`k` must be a positive integer"
        )

    a, b = r.as_integer_ratio()
    c, d = s.as_integer_ratio()

    if dir == 'left':
        return Fraction(k * a + c, k * b + d)

    return Fraction(a + k * c, b + k * d)


# A :py:func:`functools.partial` binding of :py:func:`mediant` for left-mediants.
left_mediant = functools.partial(mediant, dir="left")


# A :py:func:`functools.partial` binding of :py:func:`mediant` for right-mediants.
right_mediant = functools.partial(mediant, dir="right")


if __name__ == "__main__":      # pragma: no cover
    # Doctest the module from the project root using
    #
    #     PYTHONPATH="src" python3 -m doctest -v src/continuedfractions/lib.py
    #
    # NOTE: the doctest examples using ``decimal.Decimal`` values are based on
    #       a context precision of 28 digits
    decimal.getcontext().prec = 28
    import doctest
    doctest.testmod()



from __future__ import annotations


__all__ = [
    'ContinuedFraction',
]


# -- IMPORTS --

# -- Standard libraries --
import collections
import decimal
import functools
import math
import statistics
import typing

from decimal import Decimal
from fractions import Fraction

# -- 3rd party libraries --

# -- Internal libraries --
from continuedfractions.lib import (
    continued_fraction_rational,
    convergent,
    convergents,
    fraction_from_coefficients,
    left_mediant,
    mediant,
    remainder,
    remainders,
    right_mediant,
)


class ContinuedFraction(Fraction):
    """An object-oriented representation of a (finite) simple continued fraction.

    An implementation of simple continued fractions as Python objects and
    instances of the standard library :py:class:`fractions.Fraction` class,
    with various properties for the continued fraction, including its
    coefficients, the order, convergents, and remainders.

    The term "simple continued fraction" denotes a specific type of continued
    fraction where the fractional terms only have numerators of :math:`1`.

    Arguments can be any which are valid for creating objects of the
    :py:class:`fractions.Fraction` superclass.

    For clarification, valid arguments can be one of the following:

    * a single instance of :py:class:`numbers.Rational`, including
      :py:class:`int`, :py:class:`fractions.Fraction` or
      :py:class:`ContinuedFraction`, named or unnamed
    * a pair of  :py:class:`numbers.Rational` instances, including
      :py:class:`int`, :py:class:`fractions.Fraction` and
      :py:class:`ContinuedFraction`, named or unnamed
    * a single :py:class:`float` or :py:class:`decimal.Decimal` value
      that is not a special value such as :py:data:`math.nan`,
      ``float('inf')``, or ``Decimal('infinity')``
    * a single numeric valid string (:py:class:`str`) - validity is
      determined in the superclass by the
      :py:data:`fractions._RATIONAL_FORMAT` test
    """
    @property
    def coefficients(self) -> typing.Generator[int, None, None]:
        """:py:class:`typing.Generator`: A generator of the (ordered) sequence of coefficients of the continued fraction.

        Examples
        --------
        >>> cf = ContinuedFraction('.12345')
        >>> cf
        ContinuedFraction(2469, 20000)
        >>> tuple(cf.coefficients)
        (0, 8, 9, 1, 21, 1, 1, 5)
        """
        yield from continued_fraction_rational(self)

    @property
    def order(self) -> int:
        """:py:class:`int`: The order of the continued fraction, which is the number of its coefficients minus :math:`1`.

        Examples
        --------
        >>> cf = ContinuedFraction('.12345')
        >>> cf
        ContinuedFraction(2469, 20000)
        >>> tuple(cf.coefficients)
        (0, 8, 9, 1, 21, 1, 1, 5)
        >>> cf.order
        7
        """
        return sum(1 for coeff in self.coefficients) - 1


    @property
    def counter(self) -> collections.Counter:
        """:py:class:`collections.Counter` : A counter for the coefficients.

        Examples
        --------
        >>> cf = ContinuedFraction(928374923, 8249234)
        >>> cf.counter
        Counter({1: 6, 2: 3, 24: 2, 112: 1, 5: 1, 3: 1})
        """
        return collections.Counter(self.coefficients)

    @property
    def khinchin_mean(self) -> decimal.Decimal | None:
        """:py:class:`decimal.Decimal` or :py:data:`None`: The Khinchin mean of the continued fraction, which is defined as the geometric mean of all the tail coefficients.

        We define the Khinchin mean :math:`K_n` of a (simple) continued
        fraction :math:`[a_0; a_1, a_2, \\ldots, a_n]` as:

        .. math::

           K_n := \\sqrt[n]{a_1a_2 \\cdots a_n} = \\left( a_1a_2 \\cdots a_n \\right)^{\\frac{1}{n}}, \\hskip{3em} n \\geq 1

        This property is intended to make it easier to study the limit of
        :math:`K_n` as :math:`n \\to \\infty`.

        In the special case of integers or fractions representing integers,
        whose continued fraction representations consist of only a single
        element, a null value is returned.

        Examples
        --------
        Note that the default :py:mod:`decimal` context precision of :math:`28`
        is used in these examples.

        >>> tuple(ContinuedFraction(649, 200).coefficients)
        (3, 4, 12, 4)
        >>> ContinuedFraction(649, 200).khinchin_mean
        Decimal('5.76899828122963409526846589869819581508636474609375')
        >>> tuple(ContinuedFraction(415, 93).coefficients)
        (4, 2, 6, 7)
        >>> ContinuedFraction(415, 93).khinchin_mean
        Decimal('4.37951913988788898990378584130667150020599365234375')
        >>> tuple((ContinuedFraction(649, 200) + ContinuedFraction(415, 93)).coefficients)
        (7, 1, 2, 2, 2, 1, 1, 11, 1, 2, 12)
        >>> (ContinuedFraction(649, 200) + ContinuedFraction(415, 93)).khinchin_mean
        Decimal('2.15015313349074244086978069390170276165008544921875')
        >>> ContinuedFraction(5000).khinchin_mean
        """
        coeffs = tuple(self.coefficients)

        if self.order == 1:
            return Decimal(coeffs[-1])
        elif self.order > 1:
            return Decimal(statistics.geometric_mean(coeffs[1:]))

    @classmethod
    def from_coefficients(cls, *coeffs: int) -> ContinuedFraction:
        """Returns a :py:class:`ContinuedFraction` instance from a sequence of (integer) coefficients of a (finite) simple continued fraction.

        Invalid coefficients will trigger a :py:class:`ValueError`.

        Parameters
        ----------
        *coeffs : int
            An ordered sequence of integer coefficients of a (finite) simple
            continued fraction.

        Returns
        -------
        ContinuedFraction
            A new and fully initialised instance of :py:class:`ContinuedFraction` with
            the given sequence of coefficients.

        Raises
        ------
        ValueError
            If the given coefficients include non-integers, or where the tail
            segment contains non-positive integers.

        Examples
        --------
        Constructing a continued fraction for the rational :math:`\\frac{649}{200}` using
        the sequence of coefficients :math:`3, 4, 12, 4`.

        >>> c1 = ContinuedFraction.from_coefficients(3, 4, 12, 4)
        >>> c1
        ContinuedFraction(649, 200)

        Constructing the continued fraction of the (multiplicative) inverse :math:`\\frac{200}{649}`
        using the sequence of coefficients :math:`0, 3, 4, 12, 4`.

        >>> c2 = ContinuedFraction.from_coefficients(0, 3, 4, 12, 4)
        >>> c2
        ContinuedFraction(200, 649)

        Validation of coefficients.

        >>> ContinuedFraction.from_coefficients('0', 1)
        Traceback (most recent call last):
        ...
        ValueError: Continued fraction coefficients must be integers, and all coefficients from the 1st onwards must be positive.
        >>> ContinuedFraction.from_coefficients(0, 1, 2.5)
        Traceback (most recent call last):
        ...
        ValueError: Continued fraction coefficients must be integers, and all coefficients from the 1st onwards must be positive.
        >>> ContinuedFraction.from_coefficients(1, 0)
        Traceback (most recent call last):
        ...
        ValueError: Continued fraction coefficients must be integers, and all coefficients from the 1st onwards must be positive.
        >>> ContinuedFraction.from_coefficients(1, -1)
        Traceback (most recent call last):
        ...
        ValueError: Continued fraction coefficients must be integers, and all coefficients from the 1st onwards must be positive.
        """
        # Create a new ``ContinuedFraction`` instance from the given coefficients
        if any(not isinstance(coeff, int) or (coeff <= 0 and i > 0) for i, coeff in enumerate(coeffs)):
            raise ValueError(
                "Continued fraction coefficients must be integers, and all "
                "coefficients from the 1st onwards must be positive."
            )

        # A step to ensure uniqueness of the simple form of the continued
        # fraction - if the last element is ``1`` it can be "removed" by
        # adding it to the second last element, thereby shortening the
        # sequence by one element. The resulting simple continued
        # fraction becomes unique for the number that is represented.
        if len(coeffs) > 1 and coeffs[-1] == 1:
            coeffs = coeffs[:-2] + (coeffs[-2] + 1,)

        # Call the superclass constructor with the ``fractions.Fraction``
        # instance returned by ``lib.fraction_from_coefficients`` - this will
        # be the highest-order convergent of the simple continued
        # fraction represented by the given sequence of coefficients
        self = cls(fraction_from_coefficients(*coeffs))

        return self

    def extend(self, *new_coeffs: int) -> None:
        """Performs an in-place tail extension of the current sequence of coefficients.

        Raises a :py:class:`ValueError` if the given sequence of values is
        empty, or includes non-integers or non-positive integers.

        .. note::

           As this method performs an in-place modification of the existing/
           current instance the object ID remains the same.

        Parameters
        ----------
        new_coeffs : int
            An (ordered) sequence of new (positive) integer coefficients by
            which the tail of the existing sequence of coefficients is to be
            extended.

        Raises
        ------
        ValueError
            If the given sequence is empty or includes non-integers or non-
            positive integers.

        Examples
        --------
        >>> cf = ContinuedFraction('3.245')
        >>> cf
        ContinuedFraction(649, 200)
        >>> tuple(cf.coefficients)
        (3, 4, 12, 4)
        >>> cf.order
        3
        >>> tuple(cf.convergents)
        ((0, ContinuedFraction(3, 1)), (1, ContinuedFraction(13, 4)), (2, ContinuedFraction(159, 49)), (3, ContinuedFraction(649, 200)))
        >>> tuple(cf.remainders)
        ((3, ContinuedFraction(4, 1)), (2, ContinuedFraction(49, 4)), (1, ContinuedFraction(200, 49)), (0, ContinuedFraction(649, 200)))
        >>> cf.extend(5, 2)
        >>> cf
        ContinuedFraction(7457, 2298)
        >>> tuple(cf.coefficients)
        (3, 4, 12, 4, 5, 2)
        >>> cf.order
        5
        >>> tuple(cf.convergents)
        ((0, ContinuedFraction(3, 1)), (1, ContinuedFraction(13, 4)), (2, ContinuedFraction(159, 49)), (3, ContinuedFraction(649, 200)), (4, ContinuedFraction(3404, 1049)), (5, ContinuedFraction(7457, 2298)))
        >>> tuple(cf.remainders)
        ((5, ContinuedFraction(2, 1)), (4, ContinuedFraction(11, 2)), (3, ContinuedFraction(46, 11)), (2, ContinuedFraction(563, 46)), (1, ContinuedFraction(2298, 563)), (0, ContinuedFraction(7457, 2298)))
        >>> cf = ContinuedFraction(649, 200)
        >>> cf.extend(0, 1)
        Traceback (most recent call last):
        ...
        ValueError: The coefficients to be added to the tail must be positive integers.
        >>> cf.extend(1, -1)
        Traceback (most recent call last):
        ...
        ValueError: The coefficients to be added to the tail must be positive integers.
        """
        if not new_coeffs or any(not isinstance(e, int) or e < 1 for e in new_coeffs):
            raise ValueError(
                "The coefficients to be added to the tail must be positive "
                "integers."
            )

        coeffs = tuple(self.coefficients) + new_coeffs

        # A step to ensure uniqueness of the simple form of the continued
        # fraction - if the last of the new coefficients is ``1`` it can be
        # "absorbed" by adding it to the second last coefficient, thereby
        # shortening the sequence by one. The resulting simple continued
        # fraction becomes unique for the number that is represented.
        if len(coeffs) > 1 and coeffs[-1] == 1:
            coeffs = coeffs[:-2] + (coeffs[-2] + 1,)

        fraction = fraction_from_coefficients(*coeffs)
        self._numerator, self._denominator = fraction.as_integer_ratio()

    def truncate(self, *tail_coeffs: int) -> None:
        """Performs an in-place truncation of a contiguous trailing segment of the coefficients in the tail.

        .. note::

           As this method performs an in-place modification of the existing/
           current instance the object ID remains the same.

        Parameters
        ----------
        tail_coeffs : int
            An (ordered) sequence of (positive) integers which form a
            contiguous trailing segment of the current sequence of coefficients
            and is to be truncated.

        Raises
        ------
        ValueError
            If the tail coefficients provided are not positive integers, or do
            not form a contiguous trailing segment, or their number exceeds the
            length of the existing tail, i.e. the order of the continued
            fraction represented by the instance.

        Examples
        --------
        >>> cf = ContinuedFraction('3.245')
        >>> cf
        ContinuedFraction(649, 200)
        >>> tuple(cf.coefficients)
        (3, 4, 12, 4)
        >>> cf.order
        3
        >>> cf.counter
        Counter({4: 2, 3: 1, 12: 1})
        >>> tuple(cf.convergents)
        ((0, ContinuedFraction(3, 1)), (1, ContinuedFraction(13, 4)), (2, ContinuedFraction(159, 49)), (3, ContinuedFraction(649, 200)))
        >>> tuple(cf.remainders)
        ((3, ContinuedFraction(4, 1)), (2, ContinuedFraction(49, 4)), (1, ContinuedFraction(200, 49)), (0, ContinuedFraction(649, 200)))
        >>> cf.truncate(12, 4)
        >>> cf
        ContinuedFraction(13, 4)
        >>> tuple(cf.coefficients)
        (3, 4)
        >>> cf.order
        1
        >>> tuple(cf.convergents)
        ((0, ContinuedFraction(3, 1)), (1, ContinuedFraction(13, 4)))
        >>> tuple(cf.remainders)
        ((1, ContinuedFraction(4, 1)), (0, ContinuedFraction(13, 4)))
        >>> cf = ContinuedFraction(649, 200)
        >>> cf.truncate(4, 12)
        Traceback (most recent call last):
        ...
        ValueError: The coefficients to be truncated from the tail must consist of positive integers and form a contiguous trailing segment of the tail.
        >>> cf.truncate(3, 4, 12, 4)
        Traceback (most recent call last):
        ...
        ValueError: The coefficients to be truncated from the tail must consist of positive integers and form a contiguous trailing segment of the tail.
        """
        coeffs = tuple(self.coefficients)
        order = len(coeffs[1:])
        truncation_length = len(tail_coeffs)

        if (
            not tail_coeffs or
            any(not isinstance(coeff, int) or coeff < 1 for coeff in tail_coeffs) or
            truncation_length > order or
            coeffs[order + 1 - truncation_length:] != tail_coeffs
        ):
            raise ValueError(
                "The coefficients to be truncated from the tail must consist "
                "of positive integers and form a contiguous trailing segment "
                "of the tail."
            )

        coeffs = coeffs[:order + 1 - truncation_length]
        
        # A step to ensure uniqueness of the simple form of the continued
        # fraction - if the last coefficient is ``1`` it can be "removed" by
        # adding it to the second last coefficient, thereby shortening the
        # sequence by one. The resulting simple continued fraction becomes
        # unique for the number that is represented.
        if len(coeffs) > 1 and coeffs[-1] == 1:
            coeffs = coeffs[:-2] + (coeffs[-2] + 1,)

        fraction = fraction_from_coefficients(*coeffs)
        self._numerator, self._denominator = fraction.as_integer_ratio()

    def as_decimal(self) -> Decimal:
        """Returns the :py:class:`decimal.Decimal` value of the continued fraction.

        Returns
        -------
        decimal.Decimal
            The :py:class:`decimal.Decimal` representation of the continued fraction.

        Examples
        --------
        Note that the default :py:mod:`decimal` context precision of :math:`28`
        is used in these examples.

        >>> import math
        >>> math.pi
        3.141592653589793

        Now construct a :py:class:`ContinuedFraction` instance from it, and check the 
        :py:class:`float` value.

        >>> cf = ContinuedFraction(math.pi)
        >>> cf
        ContinuedFraction(884279719003555, 281474976710656)
        >>> cf.as_decimal()
        Decimal('3.141592653589793115997963469')
        """
        return Decimal(self.numerator) / Decimal(self.denominator)

    @functools.cache
    def convergent(self, k: int, /) -> ContinuedFraction:
        """Returns the :math:`k`-th (simple) convergent of the continued fraction.

        Given a (simple) continued fraction  :math:`[a_0;a_1,a_2,\\ldots]` the
        :math:`k`-th convergent is defined as:

        .. math::

           C_k = a_0 + \\cfrac{1}{a_1 + \\cfrac{1}{a_2 \\ddots \\cfrac{1}{a_{k-1} + \\cfrac{1}{a_k}}}}

        The result is a :py:class:`~continuedfractions.continuedfraction.ContinuedFraction` instance.
        
        If  the continued fraction is of order :math:`n` then it has exactly
        :math:`n + 1` convergents :math:`C_0,C_1,C_2,\\ldots,C_n` where
        the :math:`k`-th convergent :math:`C_k = \\frac{p_k}{q_k}` is given by
        the recurrence relation:

        .. math::
           
           \\begin{align}
           p_k &= a_kp_{k - 1} + p_{k - 2} \\\\
           q_k &= a_kq_{k - 1} + q_{k - 2},        \\hskip{3em}    k \\geq 3
           \\end{align}

        where :math:`p_0 = a_0`, :math:`q_0 = 1`, :math:`p_1 = p_1p_0 + 1`,
        and :math:`q_1 = p_1`.

        Parameters
        ----------
        k : int
            The index of the convergent, as described above.

        Returns
        -------
        ContinuedFraction
            A new :py:class:`ContinuedFraction` instance representing the :math:`k`-th
            (simple) convergent of the original continued fraction, as
            described above.

        Examples
        --------
        >>> cf = ContinuedFraction('.12345')
        >>> cf
        ContinuedFraction(2469, 20000)
        >>> cf.convergent(0)
        ContinuedFraction(0, 1)
        >>> cf.convergent(2)
        ContinuedFraction(9, 73)
        >>> cf.convergent(6)
        ContinuedFraction(448, 3629)
        >>> cf.convergent(7)
        ContinuedFraction(2469, 20000)
        """
        return self.__class__(convergent(k, *self.coefficients))

    @property
    def convergents(self) -> typing.Generator[tuple[int, ContinuedFraction], None, None]:
        """Generates an enumerated sequence of all convergents of the continued fraction.

        The convergents are generated as tuples of :py:class:`int` and
        :py:class:`~continuedfraction.continuedfraction.ContinuedFraction`
        instances, where the integers represent the indices of the convergents.

        If :math:`n` is the order of the continued fraction then :math:`n + 1`
        convergents :math:`C_0, C_1, \\ldots, C_n` are generated in that order.

        Yields
        ------
        tuple
            A tuple of convergent index (:py:class:`int`) and convergents
            (:py:class:`~continuedfractions.continuedfraction.ContinuedFraction`)
            of the continued fraction.

        Examples
        --------
        >>> cf = ContinuedFraction('3.245')
        >>> tuple(cf.convergents)
        ((0, ContinuedFraction(3, 1)), (1, ContinuedFraction(13, 4)), (2, ContinuedFraction(159, 49)), (3, ContinuedFraction(649, 200)))
        """
        yield from enumerate(map(self.__class__, convergents(*self.coefficients)))

    @property
    def even_order_convergents(self) -> typing.Generator[tuple[int, ContinuedFraction], None, None]:
        """Generates an enumerated sequence of all even-order convergents of the continued fraction.

        The convergents are generated as tuples of :py:class:`int` and
        :py:class:`~continuedfraction.continuedfraction.ContinuedFraction`
        instances, where the integers represent the indices of the convergents.

        If :math:`n` is the order of the continued fraction then only the even-
        indexed convergents :math:`C_0, C_2, C_4, \\ldots` are generated.

        Yields
        ------
        tuple
            A tuple of convergent index (:py:class:`int`) and convergents
            (:py:class:`~continuedfractions.continuedfraction.ContinuedFraction`)
            of the continued fraction.

        Examples
        --------
        >>> tuple(ContinuedFraction('3.245').even_order_convergents)
        ((0, ContinuedFraction(3, 1)), (2, ContinuedFraction(159, 49)))
        """
        yield from (t for t in self.convergents if not t[0] % 2)

    @property
    def odd_order_convergents(self) -> typing.Generator[tuple[int, ContinuedFraction], None, None]:
        """Generates an enumerated sequence of all odd-order convergents of the continued fraction.

        The convergents are generated as tuples of :py:class:`int` and
        :py:class:`~continuedfraction.continuedfraction.ContinuedFraction`
        instances, where the integers represent the indices of the convergents.

        If :math:`n` is the order of the continued fraction then only the odd-
        indexed convergents :math:`C_1, C_3, C_5, \\ldots` are generated.

        Yields
        ------
        tuple
            A tuple of convergent index (:py:class:`int`) and convergents
            (:py:class:`~continuedfractions.continuedfraction.ContinuedFraction`)
            of the continued fraction.

        Examples
        --------
        >>> tuple(ContinuedFraction('3.245').odd_order_convergents)
        ((1, ContinuedFraction(13, 4)), (3, ContinuedFraction(649, 200)))
        """
        yield from (t for t in self.convergents if t[0] % 2)

    @functools.cache
    def semiconvergent(self, k: int, m: int, /) -> ContinuedFraction:
        """Returns the :math:`m`-th semiconvergent of two consecutive convergents :math:`p_{k - 1}` and :math:`p_k` of the continued fraction.

        The integer :math:`k` must be positive and determine two consecutive
        convergents :math:`p_{k - 1}` and :math:`p_k` of a (finite, simple)
        continued fraction.

        The integer :math:`m` can be any positive integer.

        Parameters
        ----------
        k : int
            The integer :math:`k` determining two consecutive convergents
            :math:`p_{k - 1}` and :math:`p_k` of a (finite, simple) continued
            fraction
            :math:`[a_0; a_1, \\ldots, a_{k}, a_{k + 1}, \\ldots, a_n]`.

        m : int
            The index of the semiconvergent of the convergents
            :math:`p_{k - 1}` and :math:`p_k`.

        Returns
        -------
        ContinuedFraction
            The :math:`m`-th semiconvergent of the convergents
            :math:`p_{k - 1}` and :math:`p_k`.

        Raises
        ------
        ValueError
            If :math:`k` or :math:`m` are not positive integers, or :math:`k`
            is an integer that does **not** satisfy :math:`1 \\leq k \\leq n`
            where :math:`n` is the order of the (finite, simple) continued
            fraction of which :math:`p_{k - 1}` and :math:`p_k` are
            convergents.

        Examples
        --------
        >>> cf = ContinuedFraction(-415, 93)
        >>> tuple(cf.coefficients)
        (-5, 1, 1, 6, 7)
        >>> tuple(cf.convergents)
        ((0, ContinuedFraction(-5, 1)), (1, ContinuedFraction(-4, 1)), (2, ContinuedFraction(-9, 2)), (3, ContinuedFraction(-58, 13)), (4, ContinuedFraction(-415, 93)))
        >>> cf.semiconvergent(3, 1)
        ContinuedFraction(-67, 15)
        >>> cf.semiconvergent(3, 2)
        ContinuedFraction(-125, 28)
        >>> cf.semiconvergent(3, 3)
        ContinuedFraction(-183, 41)
        >>> cf.semiconvergent(3, 4)
        ContinuedFraction(-241, 54)
        >>> cf.semiconvergent(3, 5)
        ContinuedFraction(-299, 67)
        >>> cf.semiconvergent(3, 6)
        ContinuedFraction(-357, 80)
        >>> cf.semiconvergent(3, 7)
        ContinuedFraction(-415, 93)
        """
        if not isinstance(k, int) or k not in range(1, self.order + 1) or not isinstance(m, int) or m < 1:
            raise ValueError(
                "`k` and `m` must be positive integers and `k` must be an "
                "integer in the range `1..n` where `n` is the order of the "
                "continued fraction"
            )

        return self.convergent(k - 1).right_mediant(self.convergent(k), k=m)

    @functools.cache
    def remainder(self, k: int, /) -> ContinuedFraction:
        """Returns the :math:`k`-th remainder of the continued fraction.

        Given a (simple) continued fraction  :math:`[a_0;a_1,a_2,\\ldots]` the
        :math:`k`-th remainder :math:`R_k` is the (simple) continued fraction
        :math:`[a_k; a_{k + 1}, a_{k + 2}, \\ldots]`:

        .. math::

           R_k = a_k + \\cfrac{1}{a_{k + 1} + \\cfrac{1}{a_{k + 2} \\ddots }}

        where :math:`R_0` is just the original continued fraction, i.e.
        :math:`R_0 = [a_0; a_1, a_2, \\ldots]`.

        The result is a :py:class:`~continuedfractions.continuedfraction.ContinuedFraction` instance.

        Parameters
        ----------
        k : int
            The index of the remainder, as described above.

        Returns
        -------
        ContinuedFraction
            A new :py:class:`ContinuedFraction` instance representing the :math:`k`-th
            remainder of the original continued fraction, as described above.

        Examples
        --------
        >>> cf = ContinuedFraction('.12345')
        >>> cf
        ContinuedFraction(2469, 20000)
        >>> cf.remainder(0)
        ContinuedFraction(2469, 20000)
        >>> cf.remainder(2)
        ContinuedFraction(2469, 248)
        >>> cf.remainder(6)
        ContinuedFraction(6, 5)
        >>> cf.remainder(7)
        ContinuedFraction(5, 1)
        """
        return self.__class__(remainder(k, *self.coefficients))

    @property
    def remainders(self) -> typing.Generator[tuple[int, ContinuedFraction], None, None]:
        """Generates an enumerated sequence of all remainders of the continued fraction in descending order of index.

        If :math:`n` is the order of the continued fraction then there are
        :math:`n + 1` remainders :math:`R_0, R_1, \\ldots, R_n`, and the method
        generates these in reverse order :math:`R_0, R_1, \\ldots, R_n`.

        The remainders are generated as tuples of :py:class:`int`
        and :py:class:`~continuedfraction.continuedfraction.ContinuedFraction`
        instances, where the integers represent the indexes of the remainders.

        Yields
        ------
        tuple
            A tuple of remainder indices (:py:class:`int`) and remainders
            (:py:class:`~continuedfractions.continuedfraction.ContinuedFraction`)
            of the continued fraction.

        Examples
        --------
        >>> tuple(ContinuedFraction('3.245').remainders)
        ((3, ContinuedFraction(4, 1)), (2, ContinuedFraction(49, 4)), (1, ContinuedFraction(200, 49)), (0, ContinuedFraction(649, 200)))
        >>> tuple(ContinuedFraction(-415, 93).remainders)
        ((4, ContinuedFraction(7, 1)), (3, ContinuedFraction(43, 7)), (2, ContinuedFraction(50, 43)), (1, ContinuedFraction(93, 50)), (0, ContinuedFraction(-415, 93)))
        """
        coeffs = tuple(self.coefficients)
        order = len(coeffs[1:])

        yield from zip(
            reversed(range(order + 1)),
            map(self.__class__, remainders(*coeffs))
        )

    @functools.cache
    def left_mediant(self, other: Fraction, /, *, k: int = 1) -> ContinuedFraction:
        """Returns the :math:`k`-th left-mediant of the continued fraction with another rational number.
        
        For a positive integer :math:`k`, the :math:`k`-th left-mediant of two
        rational numbers :math:`r = \\frac{a}{b}` and :math:`s = \\frac{c}{d}`,
        where :math:`b, d, b + d \\neq 0`, is defined as:
        
        .. math::

           \\frac{ka + c}{kb + d}, \\hskip{3em}    k \\geq 1

        Parameters
        ----------
        other : fractions.Fraction, ContinuedFraction
            The second fraction to use to calculate the :math:`k`-th mediant with
            the first.
        
        k : int, default=1
            The order of the mediant, as defined above.        

        Returns
        -------
        ContinuedFraction
            The :math:`k`-th left-mediant of the original fraction and the second
            fraction, as a :py:class:`ContinuedFraction` instance.

        Examples
        --------
        >>> c1 = ContinuedFraction('1/2')
        >>> c2 = ContinuedFraction(3, 5)
        >>> c1, c2
        (ContinuedFraction(1, 2), ContinuedFraction(3, 5))
        >>> c1.left_mediant(c2)
        ContinuedFraction(4, 7)
        >>> c1.left_mediant(c2, k=2)
        ContinuedFraction(5, 9)
        >>> c1.left_mediant(c2, k=3)
        ContinuedFraction(6, 11)
        >>> c1.left_mediant(c2, k=100)
        ContinuedFraction(103, 205)
        >>> assert c1.left_mediant(c2, k=2) < c1.right_mediant(c2, k=2)
        >>> assert c1.left_mediant(c2, k=3) < c1.right_mediant(c2, k=3)
        >>> assert c1.left_mediant(c2, k=100) < c1.right_mediant(c2, k=100)
        """
        return self.__class__(left_mediant(self, other, k=k))

    @functools.cache
    def right_mediant(self, other: Fraction, /, *, k: int = 1) -> ContinuedFraction:
        """Returns the :math:`k`-th right-mediant of the continued fraction with another rational number.
        
        For a positive integer :math:`k`, the :math:`k`-th right-mediant of two
        rational numbers :math:`r = \\frac{a}{b}` and :math:`s = \\frac{c}{d}`,
        where :math:`b, d, b + d \\neq 0`, is defined as:
        
        .. math::

           \\frac{a + kc}{b + kd}, \\hskip{3em}    k \\geq 1

        Parameters
        ----------
        other : fractions.Fraction, ContinuedFraction
            The second fraction to use to calculate the :math:`k`-th mediant with
            the first.
        
        k : int, default=1
            The order of the mediant, as defined above.        

        Returns
        -------
        ContinuedFraction
            The :math:`k`-th right-mediant of the original fraction and the second
            fraction, as a :py:class:`ContinuedFraction` instance.

        Examples
        --------
        >>> c1 = ContinuedFraction('1/2')
        >>> c2 = ContinuedFraction(3, 5)
        >>> c1, c2
        (ContinuedFraction(1, 2), ContinuedFraction(3, 5))
        >>> c1.right_mediant(c2)
        ContinuedFraction(4, 7)
        >>> c1.right_mediant(c2, k=2)
        ContinuedFraction(7, 12)
        >>> c1.right_mediant(c2, k=3)
        ContinuedFraction(10, 17)
        >>> c1.right_mediant(c2, k=100)
        ContinuedFraction(301, 502)
        >>> assert c1.left_mediant(c2, k=2) < c1.right_mediant(c2, k=2)
        >>> assert c1.left_mediant(c2, k=3) < c1.right_mediant(c2, k=3)
        >>> assert c1.left_mediant(c2, k=100) < c1.right_mediant(c2, k=100)
        """
        return self.__class__(right_mediant(self, other, k=k))

    @functools.cache
    def mediant(self, other: Fraction, /) -> ContinuedFraction:
        """Returns the simple mediant of the continued fraction with another continued fraction.
        
        The simple mediant of two rational numbers :math:`r = \\frac{a}{b}`
        and :math:`s = \\frac{c}{d}`, where :math:`b, d, b + d \\neq 0`, is
        defined as:
        
        .. math::

           \\frac{a + c}{b + d}

        The resulting value :math:`\\frac{a + c}{b + d}` is the same as the
        1st order left- or right-mediant of :math:`r = \\frac{a}{b}`
        and :math:`s = \\frac{c}{d}`. So this method would produce the same
        result as the :py:meth:`~continuedfractions.continuedfraction.ContinuedFraction.left_mediant`
        or :py:meth:`~continuedfractions.continuedfraction.ContinuedFraction.right_mediant`
        methods where the order :math:`k` is set to :math:`1`.

        Parameters
        ----------
        other : fractions.Fraction, ContinuedFraction
            The other continued fraction.
        
        Returns
        -------
        ContinuedFraction
            The simple mediant of the original fraction and the other continued
            fraction.

        Examples
        --------
        >>> ContinuedFraction(1, 2).mediant(ContinuedFraction(3, 5))
        ContinuedFraction(4, 7)
        >>> assert ContinuedFraction(1, 2).mediant(ContinuedFraction(3, 5)) == ContinuedFraction(1, 2).left_mediant(ContinuedFraction(3, 5), k=1)
        >>> assert ContinuedFraction(1, 2).mediant(ContinuedFraction(3, 5)) == ContinuedFraction(1, 2).right_mediant(ContinuedFraction(3, 5), k=1)
        """
        return self.__class__(mediant(self, other))

    @property
    def height(self) -> int:
        """:py:class:`int` : The height of this rational number.

        Returns
        -------
        int
            The height of the fraction as a rational number represented by
            homogeneous coordinates in projective space :math:`\\mathbb{P}^1(\\mathbb{Q})`.

        Examples
        --------
        >>> ContinuedFraction(0, 1).height
        1
        >>> ContinuedFraction(-1, 2).height
        2
        >>> ContinuedFraction(3, -2).height
        3
        """
        return max(abs(self).as_integer_ratio())

    @property
    def log_height(self) -> Decimal:
        """:py:class:`decimal.Decimal` : The (natural) logarithm of the height of this rational number as defined above.

        Returns
        -------
        decimal.Decimal
            The (natural) logarithm of the height of this fraction
            as a rational number as defined above.

        Examples
        --------
        >>> ContinuedFraction(0, 1).log_height
        Decimal('0')
        >>> ContinuedFraction(-1, 2).log_height
        Decimal('0.69314718055994528622676398299518041312694549560546875')
        >>> ContinuedFraction(3, -2).log_height
        Decimal('1.0986122886681097821082175869378261268138885498046875')
        """
        return Decimal(math.log(self.height))

    def __add__(self, other, /):
        return self.__class__(super().__add__(other))

    def __radd__(self, other, /):
        return self.__class__(super().__radd__(other))

    def __sub__(self, other, /):
        return self.__class__(super().__sub__(other))

    def __rsub__(self, other, /):
        return self.__class__(super().__rsub__(other))

    def __mul__(self, other, /):
        return self.__class__(super().__mul__(other))

    def __rmul__(self, other, /):
        return self.__class__(super().__rmul__(other))

    def __truediv__(self, other, /):
        return self.__class__(super().__truediv__(other))

    def __rtruediv__(self, other, /):
        return self.__class__(super().__rtruediv__(other))

    def __floordiv__(self, other, /):
        return self.__class__(super().__floordiv__(other))

    def __rfloordiv__(self, other, /):
        return self.__class__(super().__rfloordiv__(other))

    def __divmod__(self, other, /):
        quo, rem = super().__divmod__(other)

        return self.__class__(quo), self.__class__(rem)

    def __rdivmod__(self, other, /):
        quo, rem = super().__rdivmod__(other)
        
        return self.__class__(quo), self.__class__(rem)

    def __pow__(self, other, /) -> ContinuedFraction:
        return self.__class__(super().__pow__(other))

    def __rpow__(self, other, /):
        return self.__class__(Fraction(other).__rpow__(self))

    def __pos__(self) -> ContinuedFraction:
        return self.__class__(super().__pos__())

    def __neg__(self) -> ContinuedFraction:
        """
        The negation algorithm for a finite simple continued fraction.

        The basic algorithm can be described as follows: if
        :math:`[a_0; a_1,\\ldots, a_n]` is the simple continued fraction of a
        **positive** rational number :math:`\\frac{a}{b}` of finite order
        :math:`n` then  :math:`-\\frac{a}{b}` has the simple continued
        fraction:

        .. math::

           \\begin{cases}
           [-a_0;]                                      \\hskip{3em} & n = 0 \\\\
           [-(a_0 + 1); 2]                              \\hskip{3em} & n = 1 \\text{ and } a_1 = 2 \\\\
           [-(a_0 + 1); a_2 + 1, a_3,\\ldots, a_n]      \\hskip{3em} & n \\geq 2 \\text{ and } a_1 = 1 \\\\
           [-(a_0 + 1); 1, a_1 - 1, a_2, \\ldots,a_n]   \\hskip{3em} & n \\geq 2 \\text{ and } a_1 \\geq 2
           \\end{cases}

        In applying this algorithm there is an assumption that the last
        coefficient :math:`a_n > 1`, as any simple continued fraction of type
        :math:`[a_0; a_1,\\ldots, a_{n} = 1]` can be reduced to the simple
        continued fraction :math:`[a_0; a_1,\\ldots, a_{n - 1} + 1]`.
        """
        coeffs = tuple(self.coefficients)
        n = len(coeffs)

        if n == 1:
            # Case (1) of the algorithm
            neg_coeffs = (-coeffs[0],)
        elif n == 2 and coeffs[1] == 2:
            # Case (2) of the algorithm
            neg_coeffs = (-(coeffs[0] + 1), 2)
        elif n > 1 and coeffs[1] == 1:
            # Case (3) of the algorithm
            neg_coeffs = (-(coeffs[0] + 1), coeffs[2] + 1, *coeffs[3:])
        else:
            # Case (4) of the algorithm
            neg_coeffs = (-(coeffs[0] + 1), 1, coeffs[1] - 1, *coeffs[2:])

        neg_fraction = convergent(len(neg_coeffs) - 1, *neg_coeffs)
        neg_self = self.__class__(1)
        neg_self._numerator, neg_self._denominator = neg_fraction.as_integer_ratio()

        return neg_self

    def __abs__(self) -> ContinuedFraction:
        return self.__class__(super().__abs__())


if __name__ == "__main__":      # pragma: no cover
    # Doctest the module from the project root using
    #
    #     PYTHONPATH="src" python3 -m doctest -v src/continuedfractions/continuedfraction.py
    #
    # NOTE: the doctest examples using ``decimal.Decimal`` values are based on
    #       a context precision of 28 digits.
    decimal.getcontext().prec = 28
    import doctest
    doctest.testmod()
